/* ======================================================================
 *
 *  CCCCC          vr_input.evio.c
 * CC   CC         Author(s): Bill Sherman
 * CC              Created: November 23, 2004
 * CC   CC         Last Modified: May 31, 2005
 *  CCCCC
 *
 * Code file for FreeVR inputs from the Linux Event I/O input interface.
 *
 * Copyright 2006, Bill Sherman & Friends, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/*************************************************************************

USAGE:
	This device driver currently only works on Linux systems (because
	those are where the "event I/O" interface exist -- though it may also
	be available on FreeBSD and similar systems).


	The configuration of the inputs is similar to all input devices:

	Inputs are specified with the "input" option:
		input "<name>" = "2switch(key[<number>])";
		input "<name>" = "2switch(button[<number>])";
		input "<name>" = "valuator(axis[<number>])";
		input "<name>" = "6sensor(sim6[<number>])";

*** not yet adapted for the evio interface ***
	Controls are specified in the freevrrc file:
		control "<control option>" = "2switch(button[<number>])";
		control "<control option>" = "valuator(axis[<number>])";

	Here are the available (2switch) control options for FreeVR:
	  :-(	"print_help" -- print info on how to use the input device
		"system_pause_toggle" -- toggle the system pause flag
		"print_context" -- print the overall FreeVR context data structure (for debugging)
		"print_config" -- print the overall FreeVR config data structure (for debugging)
		"print_input" -- print the overall FreeVR input data structure (for debugging)
		"print_struct" -- print the internal Evio data structure (for debugging)

		"sensor_next" -- jump to the next 6-sensor on the list
		"setsensor(<num>)" -- set the simulated sensor to a particular one
		"sensor_reset" -- reset the current 6-sensor
		"sensor_resetall" -- reset the all the 6-sensors
		"temp_valuator" -- temporarily disable 6-sensor for valuator values
		"toggle_valuator" -- toggle valuator values instead of 6-sensor
		"temp_valuator_only" -- temporarily use translation values for valuator
		"toggle_valuator_only" -- toggle whether translation is saved for valuator
		"toggle_relative" -- toggle whether movement is relative to sensor's position
		"toggle_space_limit" -- toggle whether 6-sensor can go outside working volume
		"toggle_return_to_zero" -- toggle whether return-to-zero operation is on
		"toggle_swap_transrot" -- toggle whether to swap translation and rotation inputs
	  :-(	"toggle_swap_yz" -- toggle the swapping of Y and Z translational movements

	Here are the available (valuator) control options for FreeVR:
		"set_transx" -- set the translation value for the simulated 6-sensor
		"set_transy" -- set the translation value for the simulated 6-sensor
		"set_transz" -- set the translation value for the simulated 6-sensor
		"set_azim" -- set the azimuth rotation value for the simulated 6-sensor
		"set_elev" -- set the elevation rotation value for the simulated 6-sensor
		"set_roll" -- set the roll rotation value for the simulated 6-sensor

	Here are the FreeVR configuration argument options for the Evio:
		"devname" - /dev/js<num> port Evio is connected to
			("/dev/js0" is the default)
		"restrict" - limit simulated 6-sensors to move within the working-volume
		"valuatoroverride" - produce valuator values instead of moving the 6-sensor
		"returntozero" - valuator offsets map to absolute or relative position of simulated 6-sensor
		"relativerot" - rotation of simulated 6-sensor is based on sensor's or world's coord-sys
	  	"swaptransrot" -- swap the translational and rotational inputs
		"workingvolume" - orthogonally defined volume specifying the valid range of simulated 6-sensor movement
		"transscale" - scaling factor used to tune the translational movement of a simulated 6-sensor
		"rotscale" - scaling factor used to tune the rotational movement of a simulated 6-sensor
		"valscale" - scaling factor used to tune the valuator input range
*** end of ***

HISTORY:
	23 November 2004 (Bill Sherman) -- wrote initial version by adapting
		vr_input.joydev.c code with some methods adopted from
		vr_input.xwindows.c (specifically the notion that there
		is a potential input for each key, axis, etc.).

TODO:
	- Should check to see whether some other free Unixii have the
		evio interface.

	- Make sure error handling of controls & input type mismatches is
		reasonable -- eg. trying to use a valuator control with a
		button input & vice versa.

	- When a button goes to a valuator, allow sign change operation.
		(ie. allow the sign to be specified in the "control" config line.)

	- Allow multiple inputs to same control, so for example set_roll can
		be implemented by two opposing button presses.

	- make a daemon interface

	- make the output controls (leds, sounds, forces) work -- once the
		Linux kernel can handle it

	- (perhaps) make use of incoming time data

**************************************************************************/
#if defined(__linux)	/* Can only be compiled on Linux systems { */
#if !defined(FREEVR) && !defined(TEST_APP) && !defined(CAVE)
/* Choose how this code should be compiled (ie. for CAVE, standalone, etc). */
#  define	FREEVR
#  undef	TEST_APP
#  undef	CAVE
#endif

#undef	DEBUG


#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>		/* needed for open() and fcntl() */
#ifdef __linux
#  include <sys/ioctl.h>
#  include <stdint.h>		/* needed for uint16_t type */
#  include <sys/time.h>		/* needed for struct timeval */
#  include <linux/input.h>	/* needed for Event I/O flags and masks */
	/* Some addendum types and defines that aren't including in many linux distributions */
	struct	input_devinfo {
		uint16_t	bustype;
		uint16_t	vendor;
		uint16_t	product;
		uint16_t	version;
	};

#ifdef EV_RST	/* EV_RST only seems to be defined in older versions of input.h */
	/* Older versions of linux/input.h are missing this definition */
	struct	input_absinfo {
		uint32_t	value;
		uint32_t	miniumum;
		uint32_t	maxiumum;
		uint32_t	fuzz;
		uint32_t	flat;
	};
#endif

#	define REL_RX	0x03
#	define REL_RY	0x04
#	define REL_RZ	0x05
	/* end of linux/input.h addendum */
#endif


#if defined(__CYGWIN__)
#	define KEY_MAX 0777
#	define REL_MAX 0x1F
#	define ABS_MAX 0x1F
#endif


#include "vr_debug.h"

#if defined (FREEVR)
#  include "vr_input.h"
#  include "vr_input.opts.h"
#  include "vr_parse.h"
#  include "vr_shmem.h"
#endif


#if defined(TEST_APP) || defined(CAVE)
#  define	VR_X	0
#  define	VR_Y	1
#  define	VR_Z	2
#  define	VR_W	3
#  define	VR_AZIM	0
#  define	VR_ELEV	1
#  define	VR_ROLL	2
#endif

#if defined(TEST_APP)
#  include "vr_serial.c"
#  include "vr_enums.h"
#  define	vrShmemAlloc	malloc
#  define	vrShmemAlloc0	malloc
#  define	vrAtoI		atoi
#endif


#if defined(CAVE) /* { */
#  include <math.h>
#  include "cave.h"
#  include "cave.private.h"
#  include "cave.tracker.h"
static CAVE_SENSOR_ST initial_cave_sensor = {
		0,5,0,
		0,0,0,
		0,	/*timestamp*/
		FALSE,	/*calibrated*/
		CAVE_TRACKER_FRAME
	};
#endif /* } CAVE */


	/***************************************************/
	/*** definitions for interfacing with the device ***/
	/***                                             ***/

#define	DEFAULT_DEVICE	"/dev/input/event2"	/* events 0 & 1 are usually mouse and keyboard */

/* TODO: determine whether we still want these */
/* Joydev sensitivity values */
#define TRANS_SENSITIVITY	0.1
#define ROT_SENSITIVITY		1.0
#define VALUATOR_SENSITIVITY	1.0



/****************************************************************/
/*** auxillary structure of the current data from the device. ***/
typedef struct {
		/* these are for interfacing with the hardware */
		int		fd;		/* communication file descriptor */
		fd_set		read_fds;	/* list of file descriptors for select() -- will only use one */
		char		*devfile;	/* name of device file */
		int		open;		/* flag with Evio successfully open */
		char		name[256];	/* name reported by the device */

		/* these are for internal data parsing */
		char		version[256];	/* self-reported version of the device */

#ifdef CAVE
		/* CAVE specific fields here */

#elif defined(FREEVR)
		/* FREEVR specific fields here */

		vrGenericInput	**key_inputs;		/* allocate KEY_MAX+1 */
		vrGenericInput	**rel_inputs;	/* allocate REL_MAX+1 */
		vrGenericInput	**abs_inputs;	/* allocate ABS_MAX+1 */
#define MAX_6SENSORS  16
		vr6sensor	*sensor6_inputs[MAX_6SENSORS];
		int		sim6sensor_change;	/* flag that indicates whether any of the controls to a simulated 6-sensor have changed */

		/* for the FreeVR simulated 6-sensors */
		int		active_sensor;		/* The simulated 6-sensor that is being actively controlled */
		vr6sensorConv	sensor6_options;	/* Structure of settings that affect how a 6-sensor is simulated */
		float		sim_values[6];		/* Array of 6 valuator values used to move the simulated 6-sensor */
#endif /* end library-specific fields */

#if 0 /* probably won't need this */
		/* information about the current values */
		int		num_axes;	/* number of valuators (aka axes) reported by the device */
		float		*axis;		/* array of axis values */
		int		*axis_change;	/* array of flags to indicate new valuator value */
		int		num_buttons;	/* number of buttons reported by the device */
		int		*button;	/* array of button values */
		int		*button_change;	/* array of flags to indicate new button value */
#endif

		/* information about how to filter the data for use with the VR library (FreeVR or CAVE) */
		float		scale_valuator;	/* scaling factor for valuators */
		/* TODO: determine whether any filter values/settings are necessary */

	} _EvioPrivateInfo;


/**************************************************************************/
/* arrays of strings naming all the specific inputs from the EVIO devices */
static	char	*key_names[KEY_MAX+1] = {
			"Reserved",
			"Escape",
			"1", "2", "3", "4", "5", "6", "7", "8", "9", "0",
			"Minus", "Equal", "Backspace", "Tab",
			"Q", "W", "E", "R", "T", "Y", "U", "I", "O", "P",
			"LeftBrace", "RightBrace", "Enter", "LeftCtrl",
			"A", "S", "D", "F", "G", "H", "J", "K", "L",
			"Semicolon", "Apostrophe", "Grave", "LeftShift", "BackSlash",
			"Z", "X", "C", "V", "B", "N", "M",
			"Comma", "Dot", "Slash", "RightShift",
			"KPAsterisk", "LeftAlt", "Space", "CapsLock",
			"F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9", "F10",
			"NumLock", "ScrollLock",
			"KP7", "KP8", "KP9", "KPMinus",
			"KP4", "KP5", "KP6", "KPPlus",
			"KP1", "KP2", "KP3", "KP0", "KPDot",
			"103RD", "F13", "102RD", "F11",
			"F12", "F14", "F15", "F16",
			"F17", "F18", "F19", "F20",
			"KPEnter", "RightCtl", "KPSlash", "SysRQ",
			"RightAlt", "Linefeed",
			"Home", "Up", "PageUp", "Left", "Right", "End",
			"Down", "PageDown", "Insert", "Delete",
			"Macro", "Mute", "VolumeDown", "VolumeUp", "Power",
			"KPEqual", "KPPlusMinus", "Pause",
			"F21", "F22", "F23", "F24",
			"KPComma", "LeftMeta", "RightMeta", "Compose",
			"Stop", "Again", "Props", "Undo", "Front",
			"Copy", "Open", "Paste", "Find", "Cut",
			"Help", "Menu", "Calc", "Setup", "Sleep",
			"Wakeup", "File", "Sendfile", "Deletefile",
			"Xfer", "Prog1", "Prog2", "WWW", "MsDOS",
			"Coffee", "Direction", "Cyclewindows",
			"Mail", "Bookmarks", "Computer", "Back", "Forward",
			"CloseCD", "EjectCD", "EjectCloseCD",
			"Nextsong", "PlayPause", "PreviousSong", "StopCD",
			"Record", "Rewind", "Phone", "ISO",
			"Config", "HomePage", "Refresh", "Exit",
			"Move", "Edit", "ScrollUp", "ScrollDown",
			"KPLeftParen", "KPRightParen",
			"Intl1", "Intl2", "Intl3", "Intl4",
			"Intl5", "Intl6", "Intl7", "Intl8", "Intl9",
			"Lang1", "Lang2", "Lang3", "Lang4", "Lang5",
			"Lang6", "Lang7", "Lang8", "Lang9", "0x0c7",
			"PlayCD", "PauseCD", "Prog3", "Prog4",
			"Suspend", "Close",
			"0x0ce", "0x0cf",
			"0x0d0", "0x0d1", "0x0d2", "0x0d3",
			"0x0d4", "0x0d5", "0x0d6", "0x0d7",
			"0x0d8", "0x0d9", "0x0da", "Unknown",
			"0x0dc", "0x0dd", "0x0de", "0x0df",
			"0x0e0", "0x0e1", "0x0e2", "0x0e3",
			"0x0e4", "0x0e5", "0x0e6", "0x0e7",
			"0x0e8", "0x0e9", "0x0ea", "0x0eb",
			"0x0ec", "0x0ed", "0x0ee", "0x0ef",
			"0x0f0", "0x0f1", "0x0f2", "0x0f3",
			"0x0f4", "0x0f5", "0x0f6", "0x0f7",
			"0x0f8", "0x0f9", "0x0fa", "0x0fb",
			"0x0fc", "0x0fd", "0x0fe", "0x0ff",
			"Btn:0",	/* also Btn:Misc */
			"Btn:1", "Btn:2", "Btn:3", "Btn:4",
			"Btn:5", "Btn:6", "Btn:7", "Btn:8", "Btn:9",
			"0x10a", "0x10b", "0x10c",
			"0x10d", "0x10e", "0x10f",
			"Btn:Left",	/* also Btn:Mouse */
			"Btn:Right", "Btn:Middle", "Btn:Side",
			"Btn:Extra", "Btn:Forward", "Btn:Back",
			"0x117", "0x118", "0x119", "0x11a",
			"0x11b", "0x11c", "0x11d", "0x11e", "0x11f",
			"Btn:Trigger",	/* also Btn:Joystick */
			"Btn:Thumb", "Btn:Thumb2", "Btn:Top", "Btn:Top2", "Btn:Pinkie",
			"Btn:Base", "Btn:Base2", "Btn:Base3",
			"Btn:Base4", "Btn:Base5", "Btn:Base6",
			"0x12c", "0x12d", "0x12e", "Btn:Dead",
			"Btn:A",	/* also Btn:Gamepad */
			"Btn:B", "Btn:C", "Btn:X", "Btn:Y", "Btn:Z",
			"Btn:TL", "Btn:TR", "Btn:TL2", "Btn:TR2",
			"Btn:Select", "Btn:Start", "Btn:Mode",
			"Btn:Thumbl", "Btn:Thumbr", "0x13f",
			"Btn:ToolPen",	/* also Btn:Digi */
			"Btn:ToolRubber", "Btn:ToolBrush", "Btn:ToolPencil",
			"Btn:ToolAirbrush", "Btn:ToolFinger", "Btn:ToolMouse",
			"Btn:ToolLens", "0x148", "0x149",
			"Btn:Touch", "Btn:Stylus", "Btn:Stylus2",
			"0x14d", "0x14e", "0x14f",
			"0x150", "0x151", "0x152", "0x153",
			"0x154", "0x155", "0x156", "0x157",
			"0x158", "0x159", "0x15a", "0x15b",
			"0x15c", "0x15d", "0x15e", "0x15f",
			"0x160", "0x161", "0x162", "0x163",
			"0x164", "0x165", "0x166", "0x167",
			"0x168", "0x169", "0x16a", "0x16b",
			"0x16c", "0x16d", "0x16e", "0x16f",
			"0x170", "0x171", "0x172", "0x173",
			"0x174", "0x175", "0x176", "0x177",
			"0x178", "0x179", "0x17a", "0x17b",
			"0x17c", "0x17d", "0x17e", "0x17f",
			"0x180", "0x181", "0x182", "0x183",
			"0x184", "0x185", "0x186", "0x187",
			"0x188", "0x189", "0x18a", "0x18b",
			"0x18c", "0x18d", "0x18e", "0x18f",
			"0x190", "0x191", "0x192", "0x193",
			"0x194", "0x195", "0x196", "0x197",
			"0x198", "0x199", "0x19a", "0x19b",
			"0x19c", "0x19d", "0x19e", "0x19f",
			"0x1a0", "0x1a1", "0x1a2", "0x1a3",
			"0x1a4", "0x1a5", "0x1a6", "0x1a7",
			"0x1a8", "0x1a9", "0x1aa", "0x1ab",
			"0x1ac", "0x1ad", "0x1ae", "0x1af",
			"0x1b0", "0x1b1", "0x1b2", "0x1b3",
			"0x1b4", "0x1b5", "0x1b6", "0x1b7",
			"0x1b8", "0x1b9", "0x1ba", "0x1bb",
			"0x1bc", "0x1bd", "0x1be", "0x1bf",
			"0x1c0", "0x1c1", "0x1c2", "0x1c3",
			"0x1c4", "0x1c5", "0x1c6", "0x1c7",
			"0x1c8", "0x1c9", "0x1ca", "0x1cb",
			"0x1cc", "0x1cd", "0x1ce", "0x1cf",
			"0x1d0", "0x1d1", "0x1d2", "0x1d3",
			"0x1d4", "0x1d5", "0x1d6", "0x1d7",
			"0x1d8", "0x1d9", "0x1da", "0x1db",
			"0x1dc", "0x1dd", "0x1de", "0x1df",
			"0x1e0", "0x1e1", "0x1e2", "0x1e3",
			"0x1e4", "0x1e5", "0x1e6", "0x1e7",
			"0x1e8", "0x1e9", "0x1ea", "0x1eb",
			"0x1ec", "0x1ed", "0x1ee", "0x1ef",
			"0x1f0", "0x1f1", "0x1f2", "0x1f3",
			"0x1f4", "0x1f5", "0x1f6", "0x1f7",
			"0x1f8", "0x1f9", "0x1fa", "0x1fb",
			"0x1fc", "0x1fd", "0x1fe",
			"0x1ff"			/* KEY_MAX */
		};

static	char	*rel_names[REL_MAX+1] = {
			"X", "Y", "Z",
			"RX",			/* not specified in linux/input.h" */
			"RY",			/* not specified in linux/input.h" */
			"RZ",			/* not specified in linux/input.h" */
			"HWheel", "Dial", "Wheel", "Misc",
			"0x0a", "0x0b", "0x0c", "0x0d", "0x0e",
			"0x0f"			/* REL_MAX */
		};

static	char	*abs_names[ABS_MAX+1] = {
			"X", "Y", "Z",
			"RX", "RY", "RZ",
			"Throttle", "Rudder", "Wheel", "Gas", "Brake",
			"0x0b", "0x0c", "0x0d", "0x0e", "0x0f",
			"Hat0X", "Hat0Y",
			"Hat1X", "Hat1Y",
			"Hat2X", "Hat2Y",
			"Hat3X", "Hat3Y",
			"Pressure", "Distance",
			"Tilt_X", "Tilt_Y", "Misc",
			"0x1d", "0x1e",
			"0x1f"			/* ABS_MAX */
		};



	/*********************************************/
	/*** General NON public interface routines ***/
	/*********************************************/

/******************************************************/
/* typename is used to specify a particular device among many that */
/*   share (more or less) the same protocol.  The typename is then */
/*   used to determine what specific features are available with   */
/*   this particular type of device.                               */
static void _EvioInitializeStruct(_EvioPrivateInfo *aux, char *typename)
{
#ifdef FREEVR
	/* allocate arrays based on number of possible inputs */
	aux->key_inputs =     (vrGenericInput **)vrShmemAlloc0(KEY_MAX+1 * sizeof(vrGenericInput *));
	aux->rel_inputs = (vrGenericInput **)vrShmemAlloc0(REL_MAX+1 * sizeof(vrGenericInput *));
	aux->abs_inputs = (vrGenericInput **)vrShmemAlloc0(ABS_MAX+1 * sizeof(vrGenericInput *));
#endif

	/* set input defaults */
	aux->version[0] = '\0';
	aux->name[0] = '\0';

#if 0 /* NYI */
	aux->scale_valuator = VALUATOR_SENSITIVITY;	/* set the default valuator scaling factor */
#endif

#ifndef TEST_APP /* {  (we have no simulated inputs for the test app) */
	/* set values for simulated 6-sensor */
	aux->sim_values[0] = 0.0;			/* start with no movement values in the valuators holders */
	aux->sim_values[1] = 0.0;
	aux->sim_values[2] = 0.0;
	aux->sim_values[3] = 0.0;
	aux->sim_values[4] = 0.0;
	aux->sim_values[5] = 0.0;

	aux->active_sensor = 1;				/* start with the "wand" active  -- potential bug */
	aux->active_sensor = 0;				/* TODO: go back to wand as default -- currently head (ie. delete this line) */

	aux->sensor6_options.azimuth_axis = VR_Y;	/* azimuth is about the Y axis */
	aux->sensor6_options.relative_axis = 1;		/* default to relative rotations */
	aux->sensor6_options.return_to_zero = 0;	/* default to free floating */
	aux->sensor6_options.ignore_all = 0;		/* default to joystick off */
	aux->sensor6_options.ignore_trans = 0;		/* default to joystick off */
	aux->sensor6_options.restrict_space = 1;	/* default restricted to working-volume */
	aux->sensor6_options.swap_transrot = 0;		/* default to unswapped translation & rotation */
	aux->sensor6_options.trans_scale = TRANS_SENSITIVITY;	/* set the default translational scaling */
	aux->sensor6_options.rot_scale = ROT_SENSITIVITY;	/* set the default rotational scaling */

	/* The default working-volume is that of the typical CAVE */
	aux->sensor6_options.working_volume_min[VR_X] = -5.0;
	aux->sensor6_options.working_volume_max[VR_X] =  5.0;
	aux->sensor6_options.working_volume_min[VR_Y] =  0.0;
	aux->sensor6_options.working_volume_max[VR_Y] = 10.0;
	aux->sensor6_options.working_volume_min[VR_Z] = -5.0;
	aux->sensor6_options.working_volume_max[VR_Z] =  5.0;
#endif /* } */

	/* everything else is zero'd by default */
}


/******************************************************/
static void _EvioPrintStruct(FILE *file, _EvioPrivateInfo *aux, vrPrintStyle style)
{
	char	*input_name;
	int	count;

	vrFprintf(file, "Evio device internal structure (0x%p):\n", aux);
	vrFprintf(file, "\r\tversion -- '%s'\n", aux->version);
	vrFprintf(file, "\r\tname -- '%s'\n", aux->name);
	vrFprintf(file, "\r\tfd = %d\n\tdevfile = '%s'\n\topen = %d\n",
		aux->fd,
		aux->devfile,
		aux->open);

#ifdef FREEVR /* { */
	/* print the FreeVR input objects */

#if 0 /* NYI */
	/* Only print the keyinputs that are in use -- because otherwise */
	/*   there are just too darn many of them.                       */
	vrFprintf(file, "\r\tkey/button inputs:\n");
	for (count = 0; count < KEY_MAX+1; count++) {
		if (aux->key_inputs[count] != NULL) {
			vrFprintf(file, "\r\t\tkey/button input[%d] (%-7s) = 0x%p (%s:%s)",
				count, "TODO: keyname",
				aux->button_inputs[count],
				vrInputTypeName(aux->key_inputs[count]->input_type),
				aux->key_inputs[count]->my_object->name);
		}
	}
#endif

#if 0 /* NYI */
	vrFprintf(file, "\r\tvaluator inputs:\n");
	for (count = 0; count < MAX_VALUATORS; count++) {
		vrFprintf(file, "\r\t\tvaluator_input[%d] = 0x%p", count, aux->valuator_inputs[count]);
		if (aux->valuator_inputs[count] != NULL)
			vrFprintf(file, " type: %s, name: %s\n", vrInputTypeName(aux->valuator_inputs[count]->input_type), aux->valuator_inputs[count]->my_object->name);
		else	vrFprintf(file, "\n");
	}

	vrFprintf(file, "\r\tsensor6 inputs (active = %d):\n", aux->active_sensor);
	for (count = 0; count < MAX_6SENSORS; count++) {
		vrFprintf(file, "\r\t\tsensor6_inputs[%d] = 0x%p", count, aux->sensor6_inputs[count]);
		if (aux->sensor6_inputs[count] != NULL)
			vrFprintf(file, " type: %s, name: %s\n", vrInputTypeName(aux->sensor6_inputs[count]->input_type), aux->sensor6_inputs[count]->my_object->name);
		else	vrFprintf(file, "\n");
	}
#endif

	vrFprintf(file, "\r"
			"\tsensor6_options = %d, %d, %d, %d, %d, %d, %d, %d, %.2f %.2f (%.2f--%.2f %.2f--%.2f %.2f--%.2f)\n",
			aux->sensor6_options.azimuth_axis,
			aux->sensor6_options.ignore_all,
			aux->sensor6_options.ignore_trans,
			aux->sensor6_options.tmp_ignore_trans,
			aux->sensor6_options.relative_axis,
			aux->sensor6_options.return_to_zero,
			aux->sensor6_options.restrict_space,
			aux->sensor6_options.swap_transrot,
			aux->sensor6_options.trans_scale,
			aux->sensor6_options.rot_scale,
			aux->sensor6_options.working_volume_min[0],
			aux->sensor6_options.working_volume_max[0],
			aux->sensor6_options.working_volume_min[1],
			aux->sensor6_options.working_volume_max[1],
			aux->sensor6_options.working_volume_min[2],
			aux->sensor6_options.working_volume_max[2]);
	vrFprintf(file, "\r"
			"\tsim_values = (%.2f, %.2f, %.2f,  %.2f, %.2f, %.2f)\n"
			"\tsim6sensor_change = %d\n",
			aux->sim_values[VR_X],
			aux->sim_values[VR_Y],
			aux->sim_values[VR_Z],
			aux->sim_values[VR_AZIM+3],
			aux->sim_values[VR_ELEV+3],
			aux->sim_values[VR_ROLL+3],
			aux->sim6sensor_change);
#endif /* } FREEVR */

}


/**************************************************************************/
static void _EvioPrintHelp(FILE *file, _EvioPrivateInfo *aux)
{
	vrFprintf(file, BOLD_TEXT "Sorry, Evio - print_help control not yet implemented.\n" NORM_TEXT);
}


/******************************************************/
static int _EvioReadInput(_EvioPrivateInfo *aux)
{
	char			inputs_read = 0;
#ifdef __linux /* { */
static	struct timeval		no_block = { 0 , 0 };
static	struct input_event	event;
	int			code;

	if (aux->open == 0) {
		return (0);
	}

	/* set the file descriptor list to watch the joydev device */
	FD_ZERO(&aux->read_fds);
	FD_SET(aux->fd, &aux->read_fds);

	/* loop until there is no more data on the device */
	while (select(aux->fd + 1, &aux->read_fds, NULL, NULL, &no_block)) {

		/* read the data from the device */
		read(aux->fd, &event, sizeof(event));
		code = event.code;

#if 1
		vrPrintf("Event: time = %d:%d, type = %d, code = %d, value = %d\n", event.time.tv_sec, event.time.tv_usec, event.type, event.code, event.value);
#endif

		/* parse the event */
		switch (event.type) {
#ifdef EV_RST /* not all versions of Linux have this case -- perhaps deprecated */
		case EV_RST:	/* Reset event?  currently nop */
			break;
#endif
#ifdef EV_SYN /* not all versions of Linux have this case -- perhaps new */
		case EV_SYN:	/* Sync event?  currently nop */
			break;
#endif

		case EV_KEY:	/* Key or Button event */
			/* TODO: get key/button name */
			vrDbgPrintfN(INPUT_DBGLVL, "_EvioReadInput: handling a key/button event code of 0x%02x ('%s') (input pointer = 0x%p)\n", code, key_names[code], aux->key_inputs[code]);
#ifdef FREEVR
			if (aux->key_inputs[code] != NULL) {
				/* Specifically handle mapped keys & buttons */
				switch (aux->key_inputs[code]->input_type) {
				case VRINPUT_BINARY:
					vrDbgPrintfN(INPUT_DBGLVL, "_EvioGetData: Assigning a value of %d to key 0x%x\n", event.value, code);
					vrAssign2switchValue((vr2switch *)(aux->key_inputs[code]), event.value /*, vrCurrentWallTime() */);
					break;
				case VRINPUT_VALUATOR:
					/* TODO: ... */
					/* For keys associated with valuator inputs, pressing the */
					/*   key can add some fixed, or specified amount.  Fixed  */
					/*   is easier, so we'll start with a value like 0.2.     */
					break;

				case VRINPUT_CONTROL:
					vrDbgPrintfN(INPUT_DBGLVL, "_EvioGetData: Activating a callback (0x%p) with %d to Evio key 0x%x\n",
						((vrControl *)(aux->key_inputs[code]))->callback,
						event.value, code);
					vrCallbackInvokeDynamic(((vrControl *)(aux->key_inputs[code]))->callback, 1, event.value);
					break;
				}

			} else {
				/* allow handling of "keyboard" input types here */
				/* TODO: see vr_input.xwindows.c for how to implement this */
			}
#endif
			break;

		case EV_REL:	/* Relative Axis event */
			vrDbgPrintfN(INPUT_DBGLVL, "_EvioReadInput: handling a relative-axis event code of 0x%02x ('%s') (input pointer = 0x%p)\n", code, rel_names[code], aux->rel_inputs[code]);
			break;

		case EV_ABS:	/* Absolute Axis event */
			vrDbgPrintfN(INPUT_DBGLVL, "_EvioReadInput: handling a absolute-axis event code of 0x%02x ('%s') (input pointer = 0x%p)\n", code, abs_names[code], aux->abs_inputs[code]);
			break;

		default:	/* all other reported event types */
			/* TODO: print a warning -- I don't think there should be other reported types, and if there are I'd like to know about them. */
			vrDbgPrintfN(AALWAYS_DBGLVL, "_EvioReadInput: WARNING: unknown event type of 0x%02x -- please report to library maintainer.\n", event.type);
			break;
		}
	}
#endif /* } */

	return (inputs_read);
}


/**********************************************************/
/* _EvioInitializeDevice() is called in the OPEN phase  */
/*   of input interface -- after the types of inputs have */
/*   been determined (during the CREATE phase).           */
static int _EvioInitializeDevice(_EvioPrivateInfo *aux)
{
#ifdef __linux /* { */
	long	ev_version;
#endif /* } */

	if (aux == NULL)
		return -1;

#ifdef __linux /* { */
	/* read name and parameters of device */
	ioctl(aux->fd, EVIOCGVERSION, &ev_version);
	ioctl(aux->fd, EVIOCGNAME(255), &aux->name);
	/* TODO: get all the bitmasks here */
#if 0 /* joystick stuff */
	ioctl(aux->fd, JSIOCGAXES, &aux->num_axes);
	ioctl(aux->fd, JSIOCGBUTTONS, &aux->num_buttons);
#endif

	sprintf(aux->version, "Linux EVIO Driver version 0x%x", ev_version);
#if 0 /* NYI */
	vrDbgPrintfN(INPUT_DBGLVL, "Initializing for EVIO Driver version 0x%x, name = '%s', with %d axes and %d buttons, aux->fd = %d\n", ev_version, aux->name, aux->num_axes, aux->num_buttons, aux->fd);
#endif
#else /* } { */
	sprintf(aux->version, "Non-Linux system: No Linux EVIO Driver version available");
	sprintf(aux->name, "Non-Linux system: Unable to obtain device name");
	aux->num_axes = 0;
	aux->num_buttons = 0;
#endif /* } */

#if 0 /* NYI */
	/* allocate memory for reading values from the device */
	aux->axis = vrShmemAlloc0(aux->num_axes * sizeof(float));
	aux->axis_change = vrShmemAlloc0(aux->num_axes * sizeof(int));
	aux->button = vrShmemAlloc0(aux->num_buttons * sizeof(int));
	aux->button_change = vrShmemAlloc0(aux->num_buttons * sizeof(int));
#endif

	/* read some initial data */
	_EvioReadInput(aux);

	return 0;
}


/******************************************************/
/* Translate a string name of a button (the "instance" config) into a numeric value */
static unsigned int _EvioButtonValue(char *buttonname)
{
#if 0 /* NYI */
	switch (buttonname[0]) {
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		return vrAtoI(buttonname);
		break;
	case 'a':
	case 'A':
		return Evio_BUTTONINDEX_A;
	case 'b':
	case 'B':
		return Evio_BUTTONINDEX_B;
	case 'c':
	case 'C':
		return Evio_BUTTONINDEX_C;
	case 'x':
	case 'X':
		return Evio_BUTTONINDEX_X;
	case 'y':
	case 'Y':
		return Evio_BUTTONINDEX_Y;
	case 'z':
	case 'Z':
		return Evio_BUTTONINDEX_Z;
	case 'l':
	case 'L':
		switch (buttonname[1]) {
		case '1':
			return Evio_BUTTONINDEX_L1;
		case '2':
			return Evio_BUTTONINDEX_L2;
		default:
			return -1;
		}
	case 's':	/* select and start */
	case 'r':	/* 1 and 2 */
	case 'R':
		switch (buttonname[1]) {
		case '1':
			return Evio_BUTTONINDEX_R1;
		case '2':
			return Evio_BUTTONINDEX_R2;
		default:
			return -1;
		}
	case 'S':
		return Evio_BUTTONINDEX_START;
	case 'm':
	case 'M':
		return Evio_BUTTONINDEX_MODE;
	}
#endif

	return -1;
}


/******************************************************/
/* Translate a string name of a valuator (the "instance" config) into a numeric value */
static unsigned int _EvioValuatorValue(char *valuatorname)
{
	/* skip an opening sign character */
	if (valuatorname[0] == '-' || valuatorname[0] == '+')
		valuatorname++;

#if 0 /* NYI */
	switch (valuatorname[0]) {
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		return vrAtoI(valuatorname);
		break;
	case 's':	/* slider */
	case 'S':
		return Evio_VALUATORINDEX_SLIDER;
	case 'l':	/* Left Joystick, X and Y */
	case 'L':
		switch (valuatorname[strlen(valuatorname)-1]) {
		case 'x':
		case 'X':
			return Evio_VALUATORINDEX_LEFT_X;
		case 'y':
		case 'Y':
			return Evio_VALUATORINDEX_LEFT_Y;
		default:
			return -1;
		}
	case 'r':	/* Right Joystick, X and Y */
	case 'R':
		switch (valuatorname[strlen(valuatorname)-1]) {
		case 'x':
		case 'X':
			return Evio_VALUATORINDEX_RIGHT_X;
		case 'y':
		case 'Y':
			return Evio_VALUATORINDEX_RIGHT_Y;
		default:
			return -1;
		}
	case 'd':	/* Digital Joystick, X and Y */
	case 'D':
		switch (valuatorname[strlen(valuatorname)-1]) {
		case 'x':
		case 'X':
			return Evio_VALUATORINDEX_DIGITAL_X;
		case 'y':
		case 'Y':
			return Evio_VALUATORINDEX_DIGITAL_Y;
		default:
			return -1;
		}
	}
#endif

	return -1;
}


	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/


#if defined(FREEVR) /* { */
/*****************************************************************/
/*** Functions for FreeVR access of Evio devices for input ***/
/*****************************************************************/


	/************************************/
	/***  FreeVR NON public routines  ***/
	/************************************/


/*********************************************************/
static void _EvioParseArgs(_EvioPrivateInfo *aux, char *args)
{
	float	volume_values[6];	/* for reading the working volume array */

	/*****************************************/
	/** Argument format: "devname" "=" file **/
	/**    or "devfile" "=" file            **/
	/*****************************************/
	vrArgParseString(args, "devname", &(aux->devfile));
	vrArgParseString(args, "devfile", &(aux->devfile));

	/* arguments for simulated 6-sensors */

	/***************************************************************/
	/** Argument format: "restrict" "=" { "on" | "off" | number } **/
	/***************************************************************/
	vrArgParseBool(args, "restrict", &(aux->sensor6_options.restrict_space));

	/***********************************************************************/
	/** Argument format: "valuatorOverride" "=" { "on" | "off" | number } **/
	/***********************************************************************/
	vrArgParseBool(args, "valuatoroverride", &(aux->sensor6_options.ignore_trans));

	/*******************************************************************/
	/** Argument format: "returnToZero" "=" { "on" | "off" | number } **/
	/*******************************************************************/
	vrArgParseBool(args, "returntozero", &(aux->sensor6_options.return_to_zero));

	/******************************************************************/
	/** Argument format: "relativeRot" "=" { "on" | "off" | number } **/
	/******************************************************************/
	vrArgParseBool(args, "relativerot", &(aux->sensor6_options.relative_axis));

	/*******************************************************************/
	/** Argument format: "swapTransRot" "=" { "on" | "off" | number } **/
	/*******************************************************************/
	vrArgParseBool(args, "swaptransrot", &(aux->sensor6_options.swap_transrot));

	/******************************************************************************************/
	/** Argument format: "workingVolume" "=" <float> <float> <float> <float> <float> <float> **/
	/******************************************************************************************/
	if (vrArgParseFloatList(args, "workingvolume", volume_values, 6)) {
		aux->sensor6_options.working_volume_min[VR_X] = volume_values[0];
		aux->sensor6_options.working_volume_max[VR_X] = volume_values[1];
		aux->sensor6_options.working_volume_min[VR_Y] = volume_values[2];
		aux->sensor6_options.working_volume_max[VR_Y] = volume_values[3];
		aux->sensor6_options.working_volume_min[VR_Z] = volume_values[4];
		aux->sensor6_options.working_volume_max[VR_Z] = volume_values[5];
	}

	/**********************************************/
	/** Argument format: "transScale" "=" number **/
	/**********************************************/
	if (vrArgParseFloat(args, "transscale", &(aux->sensor6_options.trans_scale))) {
		aux->sensor6_options.trans_scale *= TRANS_SENSITIVITY;
	}

	/********************************************/
	/** Argument format: "rotScale" "=" number **/
	/********************************************/
	if (vrArgParseFloat(args, "rotscale", &(aux->sensor6_options.rot_scale))) {
		aux->sensor6_options.rot_scale *= ROT_SENSITIVITY;
	}

#if 0 /* NYI */
	/********************************************/
	/** Argument format: "valScale" "=" number **/
	/********************************************/
	if (vrArgParseFloat(args, "valscale", &(aux->scale_valuator))) {
		aux->scale_valuator *= VALUATOR_SENSITIVITY;
	}
#endif

	/** TODO: other arguments to parse go here **/
}


/************************************************************/
static void _EvioGetData(vrInputDevice *devinfo)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;
	int			num_inputs;
	int			count;
	float			valuator_value;
#if 0 /* used for the simulated 6-sensor */
	float			values[6];
#endif

	num_inputs = _EvioReadInput(aux);

#if 0 /* now will be handled in _EvioReadInput() */
	if (num_inputs > 0) {

		/* handle button inputs as buttons or self-controls */
		for (count = 0; count < MAX_BUTTONS; count++) {
			if ((aux->button_change[count] != 0)) {
				if (aux->button_inputs[count] != NULL) {
					switch (aux->button_inputs[count]->input_type) {
					case VRINPUT_BINARY:
						vrAssign2switchValue((vr2switch *)(aux->button_inputs[count]), ((aux->button[count]) != 0));
						break;
					case VRINPUT_CONTROL:
						vrCallbackInvokeDynamic(((vrControl *)(aux->button_inputs[count]))->callback, 1, ((aux->button[count]) != 0));
						break;
					default:
						vrErrPrintf(RED_TEXT "_EvioGetData: Unable to handle button inputs that aren't Binary or Control inputs\n" NORM_TEXT);
						break;
					}

					aux->button_change[count] = 0;
				}
			}
		}

		/* handle valuator inputs as valuators or self-controls */
		for (count = 0; count < MAX_VALUATORS; count++) {
			if ((aux->axis_change[count] != 0)) {
				if (aux->valuator_inputs[count] != NULL) {
					valuator_value = aux->axis[count] * aux->scale_valuator * aux->valuator_sign[count];
					switch (aux->valuator_inputs[count]->input_type) {
					case VRINPUT_VALUATOR:
						vrAssignValuatorValue((vrValuator *)(aux->valuator_inputs[count]), valuator_value);
						break;
					case VRINPUT_CONTROL:
						vrCallbackInvokeDynamic(((vrControl *)(aux->valuator_inputs[count]))->callback, 1, &valuator_value);
						break;
					default:
						vrErrPrintf(RED_TEXT "_EvioGetData: Unable to handle valuator inputs that aren't Floating or Control inputs\n" NORM_TEXT);
						break;
					}
				}
				aux->axis_change[count] = 0;
			}
		}
	}
#endif

	/* handle valuator inputs to be converted to simulated 6-sensor inputs */
	if (!aux->sensor6_options.ignore_all && devinfo->num_6sensors > 0 && aux->sim6sensor_change) {
		vrAssign6sensorValueFromValuators(&(devinfo->sensor6[aux->active_sensor]), aux->sim_values, &(aux->sensor6_options), -1 /* (ie. no change to oob flag) */);
		if ((aux->sim_values[VR_X] == 0)
		   && (aux->sim_values[VR_Y] == 0)
		   && (aux->sim_values[VR_Z] == 0)
		   && (aux->sim_values[VR_AZIM+3] == 0)
		   && (aux->sim_values[VR_ELEV+3] == 0)
		   && (aux->sim_values[VR_ROLL+3] == 0)) {
			aux->sim6sensor_change = 0;
		}
	}
}


	/****************************************************************/
	/*    Function(s) for parsing Evio "input" declarations.      */
	/*                                                              */
	/*  These _Evio<type>Input() functions are called during the  */
	/*  CREATE phase of the input interface.                        */

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

/**************************************************************************/
static vrInputMatch _EvioButtonInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
        _EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;
	int			button_num;

#if 0 /* NYI */
	button_num = _EvioButtonValue(dti->instance);
	if (button_num == -1) {
		vrErrPrintf("_EvioButtonInput: Warning, button['%s'] did not match any known button\n", dti->instance);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (aux->button_inputs[button_num] != NULL)
		vrErrPrintf(RED_TEXT "_EvioButtonInput: Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	aux->button_inputs[button_num] = (vr2switch *)input;
	vrDbgPrintfN(INPUT_DBGLVL, "assigned button event of value 0x%02x to input pointer = 0x%p)\n",
		button_num, aux->button_inputs[button_num]);
#endif

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}


/**************************************************************************/
static vrInputMatch _EvioValuatorInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
        _EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;
	int			valuator_num;

#if 0 /* NYI */
	valuator_num = _EvioValuatorValue(dti->instance);
	if (valuator_num == -1) {
		vrErrPrintf("_EvioValuatorInput: Warning, valuator['%s'] did not match any known valuator\n", dti->instance);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (aux->valuator_inputs[valuator_num] != NULL)
		vrErrPrintf(RED_TEXT "_EvioValuatorInput: Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	aux->valuator_inputs[valuator_num] = (vrValuator *)input;
	if (dti->instance[0] == '-')
		aux->valuator_sign[valuator_num] = -1.0;
	else	aux->valuator_sign[valuator_num] =  1.0;
	vrDbgPrintfN(INPUT_DBGLVL, "assigned valuator event of value 0x%02x to input pointer = 0x%p)\n",
		valuator_num, aux->valuator_inputs[valuator_num]);
#endif

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}


/**************************************************************************/
static vrInputMatch _Evio6sensorInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
        _EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;
	int			sensor_num;

	sensor_num = vrAtoI(dti->instance);
	if (sensor_num < 0) {
		vrErrPrintf("_Evio6sensorInput: Warning, sensor number must be between %d and %d\n", 0, MAX_6SENSORS);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (aux->sensor6_inputs[sensor_num] != NULL)
		vrErrPrintf(RED_TEXT "_Evio6sensorInput: Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	aux->sensor6_inputs[sensor_num] = (vr6sensor *)input;

	/* the vrAssign6sensorR2Exform() function does the parsing of the next token */
	vrAssign6sensorR2Exform(aux->sensor6_inputs[sensor_num], strchr(dti->instance, ','));

	vrDbgPrintfN(INPUT_DBGLVL, "assigned 6sensor event of value 0x%02x to input pointer = 0x%p)\n",
		sensor_num, aux->sensor6_inputs[sensor_num]);

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}



	/************************************************************/
	/***************** FreeVR Callback routines *****************/
	/************************************************************/

	/************************************************************/
	/*    Callbacks for controlling the device's features.      */
	/*                                                          */

/* TODO: describe the consistent arguments of the callbacks. */
/*   ... devinfo -- same as "input", but also sometimes used to get */
/*     information about the device so it can be altered -- eg. _SpacetecSensorResetCallback()  and _XwindowsSensorNextCallback() */
/*       it is also frequently passed on to other callback routines which in */
/*       turn generally use it to access the "aux" data.  */

/************************************************************/
static void _EvioSystemPauseToggleCallback(vrInputDevice *devinfo, int value)
{
        if (value == 0)
                return;

	devinfo->context->paused ^= 1;
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: system_pause = %d.\n",
		devinfo->context->paused);
}

/************************************************************/
static void _EvioPrintContextStructCallback(vrInputDevice *devinfo, int value)
{
        if (value == 0)
                return;

        vrFprintContext(stdout, devinfo->context, verbose);
}

/************************************************************/
static void _EvioPrintConfigStructCallback(vrInputDevice *devinfo, int value)
{
        if (value == 0)
                return;

        vrFprintConfig(stdout, devinfo->context->config, verbose);
}

/************************************************************/
static void _EvioPrintInputStructCallback(vrInputDevice *devinfo, int value)
{
        if (value == 0)
                return;

        vrFprintInput(stdout, devinfo->context->input, verbose);
}

/************************************************************/
static void _EvioPrintStructCallback(vrInputDevice *devinfo, int value)
{
        _EvioPrivateInfo  *aux = (_EvioPrivateInfo *)devinfo->aux_data;

        if (value == 0)
                return;

        _EvioPrintStruct(stdout, aux, verbose);
}

/************************************************************/
static void _EvioPrintHelpCallback(vrInputDevice *devinfo, int value)
{
        _EvioPrivateInfo  *aux = (_EvioPrivateInfo *)devinfo->aux_data;

        if (value == 0)
                return;

        _EvioPrintHelp(stdout, aux);
}

/************************************************************/
static void _EvioSensorNextCallback(vrInputDevice *devinfo, int value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	if (devinfo->num_6sensors == 0) {
		vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: next sensor -- no sensors available.\n",
			aux->active_sensor);
		return;
	}

	/* set the current sensor as non-active */
	vrAssign6sensorActiveValue(((vr6sensor *)aux->sensor6_inputs[aux->active_sensor]), 0);

	/* search from the next possible sensor to the end of the list */
	do {
		aux->active_sensor++;
	} while (aux->sensor6_inputs[aux->active_sensor] == NULL && aux->active_sensor < MAX_6SENSORS);

	/* if none found (ie. we were basically already at the end of the list), then search from the beginning */
	if (aux->sensor6_inputs[aux->active_sensor] == NULL || aux->active_sensor >= MAX_6SENSORS) {
		for (aux->active_sensor = 0; aux->active_sensor < MAX_6SENSORS && aux->sensor6_inputs[aux->active_sensor] == NULL; aux->active_sensor++);
	}

	/* set the newly found sensor as active */
	vrAssign6sensorActiveValue(((vr6sensor *)aux->sensor6_inputs[aux->active_sensor]), 1);

	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: sensor6[%d] now active.\n", aux->active_sensor);
}

/* TODO: see if there is a way to call this as an N-switch */
/************************************************************/
static void _EvioSensorSetCallback(vrInputDevice *devinfo, int value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	if (value == aux->active_sensor)
		return;

	if (value < 0 || value >= MAX_6SENSORS) {
		vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: set sensor (%d) -- out of range.\n", value);
	}

	if (aux->sensor6_inputs[value] == NULL) {
		vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: set sensor (%d) -- no such sensor available.\n", value);
		return;
	}

	vrAssign6sensorActiveValue(((vr6sensor *)aux->sensor6_inputs[aux->active_sensor]), 0);
	aux->active_sensor = value;
	vrAssign6sensorActiveValue(((vr6sensor *)aux->sensor6_inputs[aux->active_sensor]), 1);

	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: 6sensor[%d] now active.\n", aux->active_sensor);
}

/************************************************************/
static void _EvioSensorSet0Callback(vrInputDevice *devinfo, int value)
{	_EvioSensorSetCallback(devinfo, 0); }

/************************************************************/
static void _EvioSensorSet1Callback(vrInputDevice *devinfo, int value)
{	_EvioSensorSetCallback(devinfo, 1); }

/************************************************************/
static void _EvioSensorSet2Callback(vrInputDevice *devinfo, int value)
{	_EvioSensorSetCallback(devinfo, 2); }

/************************************************************/
static void _EvioSensorSet3Callback(vrInputDevice *devinfo, int value)
{	_EvioSensorSetCallback(devinfo, 3); }

/************************************************************/
static void _EvioSensorSet4Callback(vrInputDevice *devinfo, int value)
{	_EvioSensorSetCallback(devinfo, 4); }

/************************************************************/
static void _EvioSensorSet5Callback(vrInputDevice *devinfo, int value)
{	_EvioSensorSetCallback(devinfo, 5); }

/************************************************************/
static void _EvioSensorSet6Callback(vrInputDevice *devinfo, int value)
{	_EvioSensorSetCallback(devinfo, 6); }

/************************************************************/
static void _EvioSensorSet7Callback(vrInputDevice *devinfo, int value)
{	_EvioSensorSetCallback(devinfo, 7); }

/************************************************************/
static void _EvioSensorSet8Callback(vrInputDevice *devinfo, int value)
{	_EvioSensorSetCallback(devinfo, 8); }

/************************************************************/
static void _EvioSensorSet9Callback(vrInputDevice *devinfo, int value)
{	_EvioSensorSetCallback(devinfo, 9); }

/************************************************************/
static void _EvioSensorResetCallback(vrInputDevice *devinfo, int value)
{
static	vrMatrix		idmat = { { 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0 } };
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;
	vr6sensor		*sensor;

	if (value == 0)
		return;

	sensor = &(devinfo->sensor6[aux->active_sensor]);
	vrAssign6sensorValue(sensor, &idmat, 0 /* , timestamp */);
	vrAssign6sensorActiveValue(sensor, -1);
	vrAssign6sensorErrorValue(sensor, 0);
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: reset 6sensor[%d].\n", aux->active_sensor);
}

/************************************************************/
static void _EvioSensorResetAllCallback(vrInputDevice *devinfo, int value)
{
static	vrMatrix		idmat = { { 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0 } };
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;
	vr6sensor		*sensor;
	int			count;

	if (value == 0)
		return;

	for (count = 0; count < MAX_6SENSORS; count++) {
		if (aux->sensor6_inputs[count] != NULL) {
			sensor = &(devinfo->sensor6[count]);
			vrAssign6sensorValue(sensor, &idmat, 0 /* , timestamp */);
			vrAssign6sensorActiveValue(sensor, (count == aux->active_sensor));
			vrAssign6sensorErrorValue(sensor, 0);
			vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: reset 6sensor[%d].\n", count);
		}
	}
}

/************************************************************/
static void _EvioTempValuatorOverrideCallback(vrInputDevice *devinfo, int value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	/* set the field to the current state of value (ie. 1 when depressed, 0 when released) */
	aux->sensor6_options.ignore_trans = value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: ignore_trans = %d.\n",
		aux->sensor6_options.ignore_trans);
}

/************************************************************/
static void _EvioToggleValuatorOverrideCallback(vrInputDevice *devinfo, int value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.ignore_trans ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: ignore_trans = %d.\n",
		aux->sensor6_options.ignore_trans);
}

/************************************************************/
static void _EvioTempValuatorOnlyCallback(vrInputDevice *devinfo, int value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	/* set the field to the current state of value (ie. 1 when depressed, 0 when released) */
	aux->sensor6_options.ignore_all = value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: ignore_all = %d.\n",
		aux->sensor6_options.ignore_all);
}

/************************************************************/
static void _EvioToggleValuatorOnlyCallback(vrInputDevice *devinfo, int value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.ignore_all ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: ignore_all = %d.\n",
		aux->sensor6_options.ignore_all);
}

/************************************************************/
static void _EvioToggleRelativeAxesCallback(vrInputDevice *devinfo, int value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.relative_axis ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: relative_axis = %d.\n",
		aux->sensor6_options.relative_axis);
}

/************************************************************/
/* TODO: this should probably also go through all the 6sensor's  */
/*   and move them to be within the allowed workspace when space */
/*   restriction is turned on.                                   */
static void _EvioToggleRestrictSpaceCallback(vrInputDevice *devinfo, int value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.restrict_space ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: restrict_space = %d.\n",
		aux->sensor6_options.restrict_space);
}

/************************************************************/
static void _EvioToggleReturnToZeroCallback(vrInputDevice *devinfo, int value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.return_to_zero ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: return_to_zero = %d.\n",
		aux->sensor6_options.return_to_zero);
}

/************************************************************/
static void _EvioToggleSwapTransRotCallback(vrInputDevice *devinfo, int value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.swap_transrot ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: swap_transrot = %d.\n",
		aux->sensor6_options.swap_transrot);
}

/************************************************************/
static void _EvioSetTransXCallback(vrInputDevice *devinfo, float *value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_X] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: set trans_x = %f.\n", *value);
}

/************************************************************/
static void _EvioSetTransYCallback(vrInputDevice *devinfo, float *value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_Y] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: set trans_y = %f.\n", *value);
}

/************************************************************/
static void _EvioSetTransZCallback(vrInputDevice *devinfo, float *value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_Z] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: set trans_z = %f.\n", *value);
}

/************************************************************/
static void _EvioSetAzimuthCallback(vrInputDevice *devinfo, float *value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_AZIM+3] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: set azimuth = %f.\n", *value);
}

/************************************************************/
static void _EvioSetElevationCallback(vrInputDevice *devinfo, float *value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_ELEV+3] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: set elevation = %f.\n", *value);
}

/************************************************************/
static void _EvioSetRollCallback(vrInputDevice *devinfo, float *value)
{
	_EvioPrivateInfo	*aux = (_EvioPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_ROLL+3] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	vrDbgPrintfN(INPUT_DBGLVL, "Evio Control: set roll = %f.\n", *value);
}



	/*************************************************/
	/*   Callbacks for interfacing with the device.  */
	/*                                               */


/************************************************************/
static void _EvioCreateFunction(vrInputDevice *devinfo)
{
	/*** List of possible inputs ***/
static	vrInputFunction	_EvioInputs[] = {
				{ "button", VRINPUT_2WAY, _EvioButtonInput },
				{ "axis", VRINPUT_VALUATOR, _EvioValuatorInput },
				{ "valuator", VRINPUT_VALUATOR, _EvioValuatorInput },
				{ "slider", VRINPUT_VALUATOR, _EvioValuatorInput },
				{ "sim6", VRINPUT_6SENSOR, _Evio6sensorInput },
				{ NULL, VRINPUT_UNKNOWN, NULL } };
	/*** List of control functions ***/
static	vrControlFunc	_EvioControlList[] = {
				/* overall system controls */
				{ "system_pause_toggle", _EvioSystemPauseToggleCallback },

				/* informational output controls */
				{ "print_context", _EvioPrintContextStructCallback },
				{ "print_config", _EvioPrintConfigStructCallback },
				{ "print_input", _EvioPrintInputStructCallback },
				{ "print_struct", _EvioPrintStructCallback },
				{ "print_help", _EvioPrintHelpCallback },

				/* simulated 6-sensor selection controls */
				{ "sensor_next", _EvioSensorNextCallback },
				{ "setsensor", _EvioSensorSetCallback },	/* NOTE: this is non-boolean */
				{ "setsensor(0)", _EvioSensorSet0Callback },
				{ "setsensor(1)", _EvioSensorSet1Callback },
				{ "setsensor(2)", _EvioSensorSet2Callback },
				{ "setsensor(3)", _EvioSensorSet3Callback },
				{ "setsensor(4)", _EvioSensorSet4Callback },
				{ "setsensor(5)", _EvioSensorSet5Callback },
				{ "setsensor(6)", _EvioSensorSet6Callback },
				{ "setsensor(7)", _EvioSensorSet7Callback },
				{ "setsensor(8)", _EvioSensorSet8Callback },
				{ "setsensor(9)", _EvioSensorSet9Callback },
				{ "sensor_reset", _EvioSensorResetCallback },
				{ "sensor_resetall", _EvioSensorResetAllCallback },

				/* simulated 6-sensor manipulation controls */
				{ "temp_valuator", _EvioTempValuatorOverrideCallback },
				{ "toggle_valuator", _EvioToggleValuatorOverrideCallback },
				{ "temp_valuator_only", _EvioTempValuatorOnlyCallback },
				{ "toggle_relative", _EvioToggleRelativeAxesCallback },
				{ "toggle_space_limit", _EvioToggleRestrictSpaceCallback },
				{ "toggle_return_to_zero", _EvioToggleReturnToZeroCallback },
				{ "toggle_swap_transrot", _EvioToggleSwapTransRotCallback },
				{ "set_transx", _EvioSetTransXCallback },
				{ "set_transy", _EvioSetTransYCallback },
				{ "set_transz", _EvioSetTransZCallback },
				{ "set_azim", _EvioSetAzimuthCallback },
				{ "set_elev", _EvioSetElevationCallback },
				{ "set_roll", _EvioSetRollCallback },

				/* other controls */
				/*   NONE   */

				/* end of the list */
				{ NULL, NULL } };

	_EvioPrivateInfo	*aux = NULL;

	/******************************************/
	/* allocate and initialize auxillary data */
	devinfo->aux_data = (void *)vrShmemAlloc0(sizeof(_EvioPrivateInfo));
	aux = (_EvioPrivateInfo *)devinfo->aux_data;
	_EvioInitializeStruct(aux, devinfo->type);

	/******************/
	/* handle options */
	aux->devfile = vrShmemStrDup(DEFAULT_DEVICE);	/* default, if no device file given */
	_EvioParseArgs(aux, devinfo->args);

	/***************************************/
	/* create the inputs and self-controls */
	vrInputCreateDataContainers(devinfo, _EvioInputs);
	vrInputCreateSelfControlContainers(devinfo, _EvioInputs, _EvioControlList);
	/* TODO: anything to do for NON self-controls?  Implement for Magellan first. */

	vrDbgPrintf("Done creating Evio inputs for '%s'\n", devinfo->name);
	devinfo->created = 1;

	return;
}


/************************************************************/
static void _EvioOpenFunction(vrInputDevice *devinfo)
{
	_EvioPrivateInfo	*aux = NULL;

	aux = (_EvioPrivateInfo *)devinfo->aux_data;

	/*******************/
	/* open the device */
#ifdef __linux /* { */
	aux->fd = open(aux->devfile, O_RDONLY);
	vrDbgPrintfN(INPUT_DBGLVL, "just opened aux->fd = %d, devfile = '%s'\n", aux->fd, aux->devfile);
	if (aux->fd < 0) {
		vrErrPrintf("(%s::_EvioOpenFunction::%d) error: "
			RED_TEXT "couldn't open joydev device %s for %s\n" NORM_TEXT,
			__FILE__, getpid(), aux->devfile, devinfo->name);
		aux->open = 0;
		sprintf(aux->version, "- unconnected Evio -");
	} else {
		if (_EvioInitializeDevice(aux) < 0) {
			vrErrPrintf("_EvioOpenFunction: "
				RED_TEXT "Warning, unable to initialize Evio '%s'.\n" NORM_TEXT, devinfo->name);
		}
		aux->open = 1;

		vrDbgPrintf("Done opening Evio input device '%s'.\n", devinfo->name);
		devinfo->ready = 1;
	}
#else /* } { ! linux */
	vrErrPrintf("_EvioOpenFunction: "
		RED_TEXT "Warning, unable to use Evio devices on non-Linux systems. ('%s' inputs exist, but without hardware).\n" NORM_TEXT, devinfo->name);
	aux->open = 0;
	devinfo->ready = 0;
#endif /* } */

	return;
}


/************************************************************/
static void _EvioCloseFunction(vrInputDevice *devinfo)
{
	_EvioPrivateInfo	*aux = NULL;

	aux = (_EvioPrivateInfo *)devinfo->aux_data;

	if (aux != NULL) {
		vrSerialClose(aux->fd);
		vrShmemFree(aux);	/* aka devinfo->aux_data */
	}

	return;
}


/************************************************************/
static void _EvioResetFunction(vrInputDevice *devinfo)
{
#if 0 /* not yet used */
	_EvioPrivateInfo	*aux = NULL;

	aux = (_EvioPrivateInfo *)devinfo->aux_data;
#endif

	return;
}


/************************************************************/
static void _EvioPollFunction(vrInputDevice *devinfo)
{
#if 0 /* not yet used */
	_EvioPrivateInfo	*aux = NULL;

	aux = (_EvioPrivateInfo *)devinfo->aux_data;
#endif

	if (devinfo->ready) {
		_EvioGetData(devinfo);
	} else {
		/* TODO: try to open the device again */
	}

	return;
}



	/******************************/
	/*** FreeVR public routines ***/
	/******************************/


/******************************************************/
void vrEvioInitInfo(vrInputDevice *devinfo)
{
	devinfo->version = (char *)vrShmemStrDup("-get from Evio device-");
	devinfo->Create = vrCallbackCreateNamed("EvioInput:Create-Def", _EvioCreateFunction, 1, devinfo);
	devinfo->Open = vrCallbackCreateNamed("EvioInput:Open-Def", _EvioOpenFunction, 1, devinfo);
	devinfo->Close = vrCallbackCreateNamed("EvioInput:Close-Def", _EvioCloseFunction, 1, devinfo);
	devinfo->Reset = vrCallbackCreateNamed("EvioInput:Reset-Def", _EvioResetFunction, 1, devinfo);
	devinfo->PollData = vrCallbackCreateNamed("EvioInput:PollData-Def", _EvioPollFunction, 1, devinfo);
	devinfo->PrintAux = vrCallbackCreateNamed("EvioInput:PrintAux-Def", _EvioPrintStruct, 0);

	vrDbgPrintfN(INPUT_DBGLVL, "vrEvioInitInfo: callbacks created.\n");
}


#endif /* } FREEVR */



	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/



#if defined(TEST_APP) /* { */

/******************************************************************/
/* Ugh, I hate globals, but I don't know a better way to get the  */
/*   aux value into the interrupt signal function exit_testapp(). */
static	int	done = 0;


/*******************************************************************/
void exit_testapp()
{
	done = 1;
}


/*******************************************************************/
/* A test program to communicate with a Evio device and print the results. */
main(int argc, char *argv[])
{
	_EvioPrivateInfo	*aux;
	int			count;

	done = 0;
	signal(SIGINT, exit_testapp);


	/******************************/
	/* setup the device structure */
	aux = (_EvioPrivateInfo *)malloc(sizeof(_EvioPrivateInfo));
	memset(aux, 0, sizeof(_EvioPrivateInfo));
	_EvioInitializeStruct(aux, "Eviosubtype");


	/*********************************************************/
	/* set default parameters based on environment variables */
	aux->devfile = getenv("EVIO_FILE");
	if (aux->devfile == NULL)
		aux->devfile = DEFAULT_DEVICE;		/* default, if no file given */


	/*****************************************************/
	/* adjust parameters based on command-line-arguments */
	/* TODO: parse CLAs for non-default serial port, baud, etc. */


	/**************************************************/
	/* open the device file and initialize the device */
	aux->fd = open(aux->devfile, O_RDONLY);
	if (aux->fd < 0) {
		fprintf(stderr, RED_TEXT "couldn't open joydev device %s\n" NORM_TEXT, aux->devfile);
		aux->open = 0;
		sprintf(aux->version, "- unconnected Evio -");
	} else {
		if (_EvioInitializeDevice(aux) < 0) {
			vrErrPrintf("main: "
				RED_TEXT "Warning, unable to initialize Evio.\n" NORM_TEXT);
		}
		aux->open = 1;
	}

	_EvioPrintStruct(stdout, aux, verbose);

	/**********************/
	/* display the output */
	while(aux->open && !done) {
		if (_EvioReadInput(aux) > 0) {
			printf("buttons: ");
			for (count = 0; count < aux->num_buttons; count++)
				printf("%d", aux->button[count]);
#if 0 /* text output of all the values */
			printf("  axes: ");
			for (count = 0; count < aux->axes; count++)
				printf("%5.1f", aux->axis[count]);
#else /* a little fancier text output */
			printf("  axes: |");
			for (count = 0; count < aux->num_axes; count++)
				printf("%5s|",
					(aux->axis[count] < -0.6 ? "-    " :
					 (aux->axis[count] < -0.3 ? " -   " :
					  (aux->axis[count] <  0.3 ? "  +  " :
					   (aux->axis[count] <  0.6 ? "   - " :
					    "    -")))));
#endif
			printf("    \r");
			fflush(stdout);
		}
	}

	printf("\n");

	/*****************/
	/* close up shop */
	if (aux != NULL) {
		close(aux->fd);
		free(aux);			/* aka devinfo->aux_data */
	}

#if 0
	_EvioCloseDevice(aux);
	vrPrintf(BOLD_TEXT "\nEvio device closed\n" NORM_TEXT);
#endif
}

#endif /* } TEST_APP */



	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/



#if defined(CAVE) /* { */


	/* ... CAVE stuff here if to also work with CAVElib */


#endif /* } CAVE */

#endif /* } __linux */

