/* ======================================================================
 *
 *  CCCCC          vr_input.fastrak.c
 * CC   CC         Author(s): Bill Sherman
 * CC              Created: April 22, 2001 (from vr_input.skeleton.c)
 * CC   CC         Last Modified: September 29, 2006
 *  CCCCC
 *
 * Code file for FreeVR inputs from the Fastrak input device.
 * And other devices that use the same (or an extended version of)
 * protocol as the Polhemus Fastrak, such as the Intersense line
 * of trackers.
 *
 * Copyright 2006, Bill Sherman & Friends, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/*************************************************************************

USAGE:
	...

	Inputs are specified with the "input" option:
	  :-(	input "<name>" = "2switch(button[<number>])";
	  :-(	input "<name>" = "valuator(joystick[<number>])";
	  :-(	input "<name>" = "valuator(camera[<number>])";
	  :-(	input "<name>" = "6sensor(receiver[<number>])";

	Controls are specified in the freevrrc file:
	  :-(	control "<control option>" = "2switch(button[<number>])";
	  ...

	Here is the available control options for FreeVR:
	  :-(	"print_help" -- print info on how to use the input device
		"system_pause_toggle" -- toggle the system pause flag
	  :-(	"print_context" -- print the overall FreeVR context data structure (for debugging)
	  :-(	"print_config" -- print the overall FreeVR config data structure (for debugging)
	  :-(	"print_input" -- print the overall FreeVR input data structure (for debugging)
	  :-(	"print_struct" -- print the internal Fastrak data structure (for debugging)
	  ...

	Here are the FreeVR configuration argument options for the Fastrak:
	  :-(	"port" - serial port Fastrak is connected to
			("/dev/input/fastrak" is the default)
			NOTE: the test app can override with FASTRAK_TTY environment variable
	  :-(	"baud" - baud rate of serial port connection
			(38400 is the default)
	  ...


NOTES on the Fastrak protocol and the InterSense IS-900:

	The initial development was done on an IS-900 reporting firmware 3.11
	Beginning with April 15, 2004, development was on an IS-900 reporting
	firmware 4.11 (the same IS-900, but probably in February of 2003, when
	it was setup on our passive stereo display wall, it got an upgrade).
	Begining in October 2004, I was using an "IS900VWT" with firmware
	version 4.2005.

	Here are some of the basic fastrack protocol commands for manual
	interaction w/ the device:

		S -- Request System Status and ID
		P -- Poll for one record of tracking info
		F -- Put in ASCII output mode
		f -- Put in Binary output mode
		C -- Place in Continuous output mode
		c -- Place in polled output mode
		U -- set output data units to inches
		u -- set output data units to centimeters
		R -- Reset reference frame to default

		l<n>\n -- station state request (on or off) (if <n> = '*' all are reported)
			[NOTE: the IS-900 IT manual makes a mistake on page 111.  It states
			that the command behaves differently when request of stations 1-4 is
			made, but in fact it does not.  It always returns the state of the
			requested station in the 4th byte, followed by 3 bytes with stations
			2, 3 & 4 (more if '*' is the station request).]
		l<n>,{0|1}\n -- disable/enable station <n>
		O<n>, ...\n -- Set the paremeters from a particular station
		O<n>\n -- Get the parameters sent from a particular station

		^K - Write current configuration into non-volitile memory
		W - Restore the system to factory defaults
		^Y - Restore the system to poweron settings
		^S - Suspend data transmission
		^Q - Resume data transmission

	Intersense additions to the protocol (NOTE: they all start with 'M'):
		ML0\n -- Turn lights off
		ML1\n -- Turn lights on
		MLB<n>\n -- Turn on light for beacon <n> (last about 7 seconds)
		MS\n -- Intersense Status Info (manufacturer specific)
		Ms<n>\n -- Station Status request [<n> = 1-c]
		Mp<n>[,<interval>] -- set/retrieve prediction interval
		MP\n -- Request Tracking status for all stations
		MCF\n -- display the configuration list of Fixed PSEs (ie. beacons)
		MCF<n>\n -- return the configuration data for a beacon
		MCF<n>,x,y,z,xn,yn,zn,IDcode\n -- set the configuration data for a beacon
		MCf<n>,IDcode\n -- remove Fixed PSE info from configuration
		MCC\n -- clear the configuration
		MCe\n -- apply the new configuration data
		MCx\n -- do not apply (cancel) the new configuration data
		ME\n -- reports all the errors in the error log
		MEC\n -- clear all the errors in the error log
		ME{0|1}\n -- disable/enable error reporting
		MLogOpen\n -- begin logging commands sent to the IS-900
		MLogClose\n -- stop logging commands sent to the IS-900
		MLogClear\n -- clear the current buffer of commands sent to the IS-900
		MLogState\n -- the current logging state
		MLogSend\n -- send the logged commands back
		MInfoSystem\n -- get info about the system (undocumented, so not all values known)
		MSystemTest1\n -- do some tests and print some information about the hardware
		MSystemTest2\n -- reports communication and tracking quality and noise

	I also discovered that if the IS-900 somehow get's stuck, then the one
	(and only) command that I found that will unstick it is "MSystemTest1\n".


	TODO: Discuss station numbers (<n>) here ...

	The configuration for the tracking device is beyond the functionality
	of this code.  The configuration is typically done once, and often
	when Intersense delivers the product and sets up the configuration.
	However, there are some systems that are more fixed, and do not require
	a theodolite, so could be configured by the local VR support staff.
	Perhaps we'll add a tool for them to do the configuration without
	using the Windows-only software provided by Intersense (ISDEMO32.exe).

	In fact, the commands to do the configuration are not that complicated.
	For example, the 3-strip setup called the "SoniFrame" can be configured
	with the following commands, So let's try again:
-------
MCC
MCF1, 0.9144,-0.6096, 0.0000, 0.00, 0.00, 1.00,5001
MCF2, 0.0000,-0.6096, 0.0000, 0.00, 0.00, 1.00,5002
MCF3,-0.9144,-0.6096, 0.0000, 0.00, 0.00, 1.00,5003
MCF4,-0.6096, 0.0000, 0.0000, 0.00, 0.00, 1.00,5004
MCF5, 0.0000, 0.0000, 0.0000, 0.00, 0.00, 1.00,5005
MCF6, 0.6096, 0.0000, 0.0000, 0.00, 0.00, 1.00,5006
MCF7, 0.9144, 0.6096, 0.0000, 0.00, 0.00, 1.00,5007
MCF8, 0.0000, 0.6096, 0.0000, 0.00, 0.00, 1.00,5008
MCF9,-0.9144, 0.6096, 0.0000, 0.00, 0.00, 1.00,5009
MCe
MCF
-------
	And then a ^K to write the data to non-volitile memory.


	Polhemus Patriot protocol:

		The Patriot device is somewhat similar to the original
		Fastrak, but there are some significant differences.

		First, when the Patriot is powered up, it sends the string
		"Patriot Ready!\n\n" [and it's possible it actually sends
		that on both the TX and RX lines, since the information seems
		to get out when a null-modem is in the serial cable, but
		then information doesn't seem to go back.

		Second, the case of the commands seems to be ignored, so
		commands like "F" and "f" now take an argument to specify
		the format, rather than using the case of the command.
		Fortunately, the output formats (and "O" command) seem to
		work more or less the same (at least for ASCII, binary was
		not checked).

		Also, it seems that all commands (except "P") require a
		carriage return in order to execute them.

		So here are the key commands for the Patriot:

			P -- Poll for one record of tracking info (and if
				in continuous mode, switch to polled mode)
			F\n -- report output mode
			F0\n -- Put in ASCII output mode
			F1\n -- Put in Binary output mode
			C\n -- Place in Continuous output mode
			U\n -- report output data units
			U0\n -- set output data units to inches
			U1\n -- set output data units to centimeters
			^V{,0,1,2,3}\n -- report version/serial information
				[0 is transmitter ID, 1 & 2 are station IDs,
				blank or 3 reports system version information]

		Here are some other possibly useful commands:
			^T[0]\n -- report/clear startup errors
			^E{,0,1}\n -- report/set echo mode (0 - off, 1 - on)
			^U...\n -- report/set station state(s)
				[similar to the Fastrak "l" command]
			^X...\n -- report/set internally stored configurations
			R -- Reset reference frame to default (or perhaps this
				is ^R -- there seems to be a discrepancy between
				the manual and the response from the unit)
			X...\n -- set location filter parameters
			Y...\n -- set attitude filter parameters
			@B...\n -- USB buffering mode -- didn't work as
				documented for me (perhaps because USB cable
				not connected?)
			^Y -- go through power-up initialization (takes a
				couple seconds)
			^Z{,0,1,2,3} -- report configuration settings
				[blank is current, 0 is factory default
				settings, 1-3 are user defined config settings.]

		And here are some significant missing commands:
			S -- Request System Status and ID

		And as far as the output format (ie. "O") values go, there are
		some the same, and some different.  The ones that are the same
		are:
			0 - ASCII space
			1 - ASCII CR/LF
			2 - X,Y,Z cartesian coordinate
			4 - Az,El,Rl Euler orientation angles
		which are the ones FreeVR typically uses.  The other value to
		note is:
			10 (patriot) -- stylus value
				[and currently (09/22/2006), I cannot find a way to
				determine which button on the Fledermaus controller
				is pressed -- all report "1"]
				[and after some inquiries, the current sentiment is
				that all the buttons are mapped to the same physical
				input, so there really is only one button to build
				the interface around.]
		Also, there seems to be a consistency in that the maximumn parameter
		value is 66 -- even though many of them are undefined.

HISTORY:
	22 April 2001 (Bill Sherman)
		I copied vr_input.skeleton.c and made the intial changes to
		  convert code to that of a specific device.

	3 May 2001 (Bill Sherman)
		I don't have this device working yet, but I'm catching up all
		  the input devices to a few minor changes made to the general
		  vr_input.skeleton.c format, and this one is no exception.

	6 May 2001 (Bill Sherman) (mid-mod)
		I brought some of the concepts from the FOB code into this
		  file.  Primarily the concept of a data structure for each
		  unit of the overall device.
		I also began writing a general decoder routine for all incoming
		  messages.

	14 June 2001 (Bill Sherman) (mid-mod)
		Got basics (very) working.  I.e. it reads the system id, and
		  knows that it got an output list record or a data record.

	20 May 2002 (Bill Sherman)
		A few modifications to keep up with changes to the library such
		  as the renaming of input types (sensor6 to 6sensor, etc), and
		  some #define'd values (X to VR_X, Y to VR_Y, etc).

	11 September 2002 (Bill Sherman)
		Moved the control callback array into the _FastrakFunction()
		callback.

	21-23 April 2003 (Bill Sherman)
		Updated to new input matching function code style.  Changed
		  "opaque" field to "aux_data".  Split _FastrakFunction() into
		  5 functions.  Added new vrPrintStyle argument to
		  _FastrakPrintStruct() for the sake of the new "PrintAux"
		  callback.

	3 June 2003 (Bill Sherman)
		Now include "vr_enums.h" for the TEST_APP code.
		Added the address of the auxillary data to the printout.
		Added the "system_pause_toggle" control callback.
		Added comments classifying the contols.

	15 April 2004 (Bill Sherman)
		After no real work since June of 2001 (apart from some basic
		  general library changes), I'm back to working on the IS-900
		  tracker input device.

	25 October 2004 (Bill Sherman)
		Again after a bout of no real work, I finally got around to
		  actually reading data in a useful fashion.  However, a lot
		  of stuff was hardcoded for a particular setup so I could
		  have some FreeVR apps running at the DRI demo days.  I
		  managed to get the test application to function minimally
		  (copying over the ASCII display routine from vr_input.asc_fob.c),
		  and began to work on the FreeVR version, but not enough of
		  the skeleton contained the necessary code, so more stuff
		  was copied from the Flock of Birds driver code.

	27 October 2004 (Bill Sherman, in Las Vegas with the FakeSpace ROVR)
		I worked in Las Vegas with the Fakespace ROVR on loan.  I
		  figured out how to parse the "O<n>" responses from the
		  IS-900, and got pretty close with the reults, but not enough
		  time to getting really working before the demos.

	18-29 August 2005 (Bill Sherman) (mid-mod)
		Focused on getting the "fastraktest" program functioning
		  properly (and the line-rendering version at that).  In the
		  process added several debugging print statements, but have
		  been #if'd out now that things have begun working.
		Added the line-rendering output routine.

	2-8 September 2005 (Bill Sherman) (mid-mod)
		Discovered a decoding bug, and finally discovered and corrected
		  the fatal flaw.  (Search for the word "robust" for how to
		  recreate errors, so that the rest of the code can be made
		  more robust.)

	12-14 September 2005 (Bill Sherman) (mid-mod)
		I'm fairly happy with the fastraktest implementation, and now I
		  need to get the FreeVR interface completed.

	18-19 September 2005 (Bill Sherman)
		Implemented the FreeVR interface.

	21 October 2005 (Bill Sherman)
		Added a new time-plot of tracker quality compile-time option to
		the fastraktest application.

	6 March 2006 (Bill Sherman)
		Fixed some type casts to make the code cleaner (and compile
		  without warnings on pickier architectures).

	22-28 September 2006 (Bill Sherman)
		Implemented the Polhemus Patriot protocol (which is somewhat similar
		  to the standard Fastrak protocol, but not really -- vaguely
		  familiar is probably a better way to describe it.  In any event,
		  however, in the end I decided to keep it merged within this
		  source file.


TODO:
	- The rest of the implementation:
		- are there commands to get the devices Global status, Error
			status, and Unit Status?
		- go into a stream mode if specified
			NOTE: with current protocol, we can't know the tracking "letter"
				when in stream mode -- since that has to be polled separately.
		- go into a binary reporting mode if specified
		- separate out some of the IS-900 stuff
		- do some error checking
		- should parse for the error code (e.g. "2 E*ERROR*O:,40,2,4,1*ERROR* EC -1*PS1  *FL1 *ST0") -- in this case caused by the ':' character sent as a station number.
			[8/30/05: which would have been caused by a unit-10, and
			not checking to make sure the correct ASCII value was
			given to address the 10th (or 11th or 12th) reciever
			units.]

	- test whether "\n\r" is the proper terminator for Fastrak & IS-900 (the Patriot
		seems to prefer only "\r" to terminate the commands.  Thus, many of
		the commands particular to Polhemus devices (such as Hemisphere)
		have only been tested on the Patriot.

	- PARTIAL: test the delay values -- current values were determined for
		the FOB [I've reduced the values, and it still works]

	- search for the word "robust", and deal with related issues.

	- should be able to treat loss of tracking on any/all sensor(s)
		as a button press to the application (or used as a
		control to other parts of FreeVR to make a sound or
		whatever).  The front panel reports three tracker conditions:
		'X', 'L' and 'T'.  'T' meens good quality tracking, and I think
		'L' means low quality tracking, and I'm guessing that 'X' means
		no tracker (perhaps because there is no communication with a
		wireless device).  Also need to see if there is any other flags
		that indicate lack of wireless communication.

		Another (perhaps duplicate) thing to do is have an n-sensor with
		the three values.

	- PARTIAL: add to test program, an option to display the beacon
		information off to the side of the rendering (perhaps even a
		"graphical" display).

	- write a companion program for configuring an IS-900 device (ie.
		entering the beacon locations and sending to device.)

	- add envvars and command line arguments for which receiver has
		buttons/joystick, and other configuration options (eg. stream
		mode).

	- write an IS900 ethernet protocol communications option (vs. the RS-232)

	- get this code tested on an actual Polhemus Fastrak system

	- determine whether there is a way to query whether a particular
		receiver has buttons.


**************************************************************************/
#if !defined(FREEVR) && !defined(TEST_APP) && !defined(CAVE)
/* Choose how this code should be compiled (ie. for CAVE, standalone, etc). */
#  define	FREEVR
#  undef	TEST_APP
#  undef	CAVE
#endif

#define	DEBUG_PRINT
#undef COMM_DEBUG	/* define this to add a lot of output for debuging communications w/ the device */


#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>

#include "vr_serial.h"
#include "vr_debug.h"
#include "vr_utils.h"


#if defined (FREEVR)
#  include "vr_input.h"
#  include "vr_input.opts.h"
#  include "vr_parse.h"
#  include "vr_shmem.h"
#endif


#if defined(TEST_APP) || defined(CAVE)
#  define	VR_X	0
#  define	VR_Y	1
#  define	VR_Z	2
#  define	VR_AZIM	0
#  define	VR_ELEV	1
#  define	VR_ROLL	2
#endif

#if defined(TEST_APP)
#  include <stdlib.h>		/* needed for getenv() & atoi() */
#  include "vr_serial.c"
#  include "vr_utils.c"
#  include "vr_enums.h"
#endif


/*** define other library stuff here if desired ***/


/*** local defines ***/

#define STATION_NUM2CHAR(s)	('0' + ((char)(s)) + (((char)(s) > (char)9) ? (char)7 : (char)0))
#define STATION_CHAR2NUM(s)	(int)((((char)(s)) - '0') - (((char)(s) > '9') ? (char)7 : (char)0))

#if 0
#define	DEFAULT_PORT	"/dev/input/fastrak"
#elif 0
#define	DEFAULT_PORT	"/dev/ttyS0"
#elif 0
#define	DEFAULT_PORT	"/dev/ttyIOC4/1"
#else
#define	DEFAULT_PORT	"/dev/ttyUSB0"
#endif
#define	DEFAULT_BAUD	115200

#define	BUFSIZE		1024


	/******************************************************************/
	/*** definitions for interfacing with a Fastrak protocol device ***/
	/***                                                            ***/

/* some of the timing delay values */
#if 1
#define FASTRAK_EXAM_DELAY		 10000		/* Warning: only tested on an IS-900 VETracker */
#define FASTRAK_SET_DELAY		  5000		/* Warning: only tested on an IS-900 VETracker */
#else
/* these are for testing Patriot via serialspy */
#define FASTRAK_EXAM_DELAY		 1000000
#define FASTRAK_SET_DELAY		  500000
#endif

/* Fastrak button bit-masks & indicies */
#define Fastrak_BUTTON_x1x	0x01
#define Fastrak_BUTTON_x2x	0x02
#define Fastrak_BUTTON_x3x	0x04
#define Fastrak_BUTTON_x4x	0x08
#define Fastrak_BUTTON_x5x	0x10
#define Fastrak_BUTTON_x6x	0x20
#define Fastrak_BUTTON_x7x	0x40

#define Fastrak_BUTTONINDEX_x1x	0x00
#define Fastrak_BUTTONINDEX_x2x	0x01
#define Fastrak_BUTTONINDEX_x3x	0x02
#define Fastrak_BUTTONINDEX_x4x	0x03
#define Fastrak_BUTTONINDEX_x5x	0x04
#define Fastrak_BUTTONINDEX_x6x	0x05
#define Fastrak_BUTTONINDEX_x7x	0x06

#define Fastrak_VALUATORINDEX_X	 0
#define Fastrak_VALUATORINDEX_Y	 1

/* some possible manufacturer values */
#define	MANU_POLHEMUS	1
#define	MANU_INTERSENSE	2

/* some possible models */
#define MODEL_FASTRAK	1
#define MODEL_PATRIOT	2
#define MODEL_LIBERTY	3
#define MODEL_IS900	4


/*******************************/
/* Device communications codes */
typedef	enum {
		FASTRAK_SYSSTAT,

		FASTRAK_ASCIIMODE,
		FASTRAK_BINARYMODE,

		FASTRAK_STREAMMODE,
		FASTRAK_POLLMODE,
		FASTRAK_POLL,

		FASTRAK_SET_INCHES,
		FASTRAK_SET_CENTIMETERS,

		FASTRAK_GET_ALLRECEIVERS,

		/* Some commands specific to the Polhemus Patriot */
		PATRIOT_POLLMODE,
		PATRIOT_ASCIIMODE,
		PATRIOT_SET_INCHES,
		PATRIOT_SYSSTAT,
		PATRIOT_GET_ALLRECEIVERS,

		FASTRAK_SET_HEMISPHERE_FORE,
		FASTRAK_SET_HEMISPHERE_AFT,
		FASTRAK_SET_HEMISPHERE_RIGHT,
		FASTRAK_SET_HEMISPHERE_LEFT,
		FASTRAK_SET_HEMISPHERE_LOWER,
		FASTRAK_SET_HEMISPHERE_UPPER,

		FASTRAK_SET_OUTPUT_ANGLES,
		FASTRAK_SET_OUTPUT_MATRIX,
		FASTRAK_SET_OUTPUT_POSITION,
		FASTRAK_SET_OUTPUT_POSITION_ANGLES,
		FASTRAK_SET_OUTPUT_POSITION_ANGLES_BJ,
		FASTRAK_SET_OUTPUT_POSITION_MATRIX,
		FASTRAK_SET_OUTPUT_POSITION_MATRIX_BJ,
		FASTRAK_SET_OUTPUT_POSITION_QUATERNION,
		FASTRAK_SET_OUTPUT_POSITION_QUATERNION_BJ,
		FASTRAK_SET_OUTPUT_QUATERNION,
		PATRIOT_SET_OUTPUT_POSITION_ANGLES,
		PATRIOT_SET_OUTPUT_POSITION_ANGLES_B,

		FASTRAK_GET_STATION_PARAMS,

		IS900_GET_TUNITS,
		IS900_GET_TRACKING_STATUS,
		IS900_SET_LIGHTS,

		FASTRAK_LASTCOMMAND		/* used to check bounds on array */
	} _FastrakCommand;


/**********************************************/
/* _FastrakComandInfo structure associates the command names with   */
/*    the Fastrak protocol command byte sequence, (incl. the number */
/*    of bytes, a post-command delay, and a string with the name.   */
typedef struct {
		char	name[128];
		int	len;
		int	us_delay;
		char	msg[32];
	} _FastrakCommandInfo;

	/* NOTE: these are not proper C-strings, in that '\0' is a valid character,   */
	/*   So some of the character arrays require the individual characters to be  */
	/*   specified as an array of chars, rather than a C-string -- in particular, */
	/*   strings that contain the NUL value.                                      */

	/* NOTE: (also) that this list is not comprehensive.  It contains all the     */
	/*   commands that are deemed (or were once deemed) necessary for normal      */
	/*   operation of the device.                                                 */
static	_FastrakCommandInfo	_FastrakMsgList[] = {
		{ "FASTRAK_SYSSTAT",		1,	FASTRAK_EXAM_DELAY,	"S" },	/* pg 93 */

		{ "FASTRAK_ASCIIMODE",		1,	FASTRAK_SET_DELAY,	"F" },	/* pg 94 */
		{ "FASTRAK_BINARYMODE",		1,	FASTRAK_SET_DELAY,	"f" },	/* pg 94 */
		{ "FASTRAK_STREAMMODE",		1,	FASTRAK_SET_DELAY,	"C" },	/* pg 90 */
		{ "FASTRAK_POLLMODE",		1,	FASTRAK_SET_DELAY,	"c" },	/* pg 90 */
		{ "FASTRAK_POLL",		1,	FASTRAK_SET_DELAY,	"P" },	/* pg 90 */

		{ "FASTRAK_SET_INCHES",		1,	FASTRAK_SET_DELAY,	"U" },	/* pg 93 */
		{ "FASTRAK_SET_CENTIMETERS",	1,	FASTRAK_SET_DELAY,	"u" },	/* pg 93 */

		{ "FASTRAK_GET_ALLRECEIVERS",	3,	FASTRAK_EXAM_DELAY,	"l*\n"}, /* pg ? */

		{ "PATRIOT_POLLMODE",		1,	FASTRAK_SET_DELAY,	"p" },	/* either 'P' or 'p' works, I do this for debugging while monitoring the stream */
		{ "PATRIOT_ASCIIMODE",		3,	FASTRAK_SET_DELAY,	"F0\r" },
		{ "PATRIOT_SET_INCHES",		3,	FASTRAK_SET_DELAY,	"U0\r" },
		{ "PATRIOT_SYSSTAT",		2,	FASTRAK_SET_DELAY,	"\026\r" },	/* ^V */
		{ "PATRIOT_GET_ALLRECEIVERS",	3,	FASTRAK_SET_DELAY,	"\0251\r" },	/* ^U */

	/* TODO: the hemisphere parameters have not been tested, also station number needs to be set in the string. */
		/* NOTE: hemisphere commands only tested on Patriot -- uses just "\r" as terminator rather than "\n\r" */
		{ "FASTRAK_SET_HEMISPHERE_FORE",   9,	FASTRAK_SET_DELAY,	"H1,1,0,0\r" },
		{ "FASTRAK_SET_HEMISPHERE_AFT",   10,	FASTRAK_SET_DELAY,	"H1,-1,0,0\r" },
		{ "FASTRAK_SET_HEMISPHERE_RIGHT" , 9,	FASTRAK_SET_DELAY,	"H1,0,1,0\r" }, /* pg 100 */
		{ "FASTRAK_SET_HEMISPHERE_LEFT",  10,	FASTRAK_SET_DELAY,	"H1,0,-1,0\r" },
		{ "FASTRAK_SET_HEMISPHERE_LOWER",  9,	FASTRAK_SET_DELAY,	"H1,0,0,1\r" },
		{ "FASTRAK_SET_HEMISPHERE_UPPER", 10,	FASTRAK_SET_DELAY,	"H1,0,0,-1\r" },

		/* NOTE: the "40" parameter in the following output option commands are actually IS900 sepecific */

		{ "FASTRAK_SET_OUTPUT_ANGLES",	11,	FASTRAK_SET_DELAY,	"O1,40,4,1\n\r" },	/* pg 95-99 */
		{ "FASTRAK_SET_OUTPUT_MATRIX",	15,	FASTRAK_SET_DELAY,	"O1,40,5,6,7,1\n\r" },
		{ "FASTRAK_SET_OUTPUT_POSITION",11,	FASTRAK_SET_DELAY,	"O1,40,2,1\n\r" },
		{ "FASTRAK_SET_OUTPUT_POSITION_ANGLES",13,FASTRAK_SET_DELAY,	"O1,40,2,4,1\n\r" },
		{ "FASTRAK_SET_OUTPUT_POSITION_ANGLES_BJ",19,FASTRAK_SET_DELAY,	"O1,40,2,4,22,23,1\n\r" },
		{ "FASTRAK_SET_OUTPUT_POSITION_MATRIX",17,FASTRAK_SET_DELAY,	"O1,40,2,5,6,7,1\n\r" },
		{ "FASTRAK_SET_OUTPUT_POSITION_MATRIX_BJ",23,FASTRAK_SET_DELAY,	"O1,40,2,5,6,7,22,23,1\n\r" },
		{ "FASTRAK_SET_OUTPUT_POSITION_QUATERNION",14,FASTRAK_SET_DELAY,"O1,40,2,11,1\n\r" },
		{ "FASTRAK_SET_OUTPUT_POSITION_QUATERNION_BJ",20,FASTRAK_SET_DELAY,"O1,40,2,11,22,23,1\n\r" },
		{ "FASTRAK_SET_OUTPUT_QUATERNION",12,	FASTRAK_SET_DELAY,	"O1,40,11,1\n\r" },
		{ "PATRIOT_SET_OUTPUT_POSITION_ANGLES",9,FASTRAK_SET_DELAY,	"O1,2,4,1\r" },
		{ "PATRIOT_SET_OUTPUT_POSITION_ANGLES_B",12,FASTRAK_SET_DELAY,	"O1,2,4,10,1\r" },

		{ "FASTRAK_GET_STATION_PARAMS", 3,	FASTRAK_SET_DELAY,	"O1\r" },

		/********************************/
		/* InterSense specific commands */
		{ "IS900_GET_TUNITS",		4,	FASTRAK_EXAM_DELAY,	"MCF\n" },
		{ "IS900_GET_TRACKING_STATUS",	3,	FASTRAK_EXAM_DELAY,	"MP\n" },
		{ "IS900_SET_LIGHTS",		4,	FASTRAK_SET_DELAY,	"ML0\n" }, /* pg 103 */

		{ "FASTRAK_LASTCOMMAND",	0,	0,			"" }
	};



/******************************************************************/
/*** auxillary structures of the current data from the Fastrak. ***/

	/* _FastrakUnit: contains data for a single unit of the Fastrak system */
typedef struct {
		unsigned char	state;		/* state byte of the sensor '0'=off, '1' = on */
		unsigned char	status[3];	/* from intersense status ("Ms/n") */
		char		toreport[128];	/* what the system reports it will report ("O<n>/n") */
		int		toreport_len;	/* number of bytes expected for each report */
		int		toreport_type[16];/* an array of what values to expect */
		int		toreport_num;	/* the number of items in the toreport_type array */
		int		read_count;	/* the number of times this device has reported values */
		float		time_stamp;	/* incoming time data from sensor [TODO: use this] */
		float		data_pos[32];	/* incoming position data of the sensor */
		int		buttons;	/* incoming button data of the sensor */
		int		joystick[2];	/* incoming joystick data of the sensor */
		int		has_button;	/* whether button/joystick is attached to this unit */
		unsigned char	tracking_quality;/* incoming tracking quality of the sensor */
		char		tracking_status;/* incoming tracking status of the sensor (from IS900 "MP" command) */
		int		tracking_meas;	/* number of range measurements received this cycle ("MP") */
		int		tracking_reject;/* number of range measurements rejected this cycle ("MP") */

		/* other possibilities: */
		/*	- alignment reference frame: 9 floats */
		/*	- boresight reference angles: 3 floats */
		/*	- hemisphere parames: 3 floats */
		/*	- tip offset values: 3 floats */
		/*	- location operational envelop: 6 floats */
		/*	- prediction interval: 1 integer */
		/*	- sensitivity level: 1 integer (1-5) */

	} _FastrakUnit;

	/* _FastrakTransUnit: contains data for a single transmitter unit */
	/*   (e.g. on the IS-900 system each beacon is a separate unit)   */
typedef struct {
		int		assigned;	/* boolean flag to indicate the existance of this unit */
		float		position[6];	/* location and directional normal */
		int		pse_id;		/* position-sensing-element id number (usually 1..n) */
		int		hardware_id;	/* the hardware-encoded ID for the element (soniDisc) */
	} _FastrakTransUnit;

typedef struct {
		/* these are for interfacing with the hardware */
		int		fd;		/* was commhandle */
		char		*port;		/* name of serial port */
		int		baud_enum;	/* communication rate as an enumerated value */
		int		baud_int;	/* communication rate as the real value */
		int		open;		/* flag with Fastrak successfully open */

		/* these are for internal data parsing */
		unsigned char	buf[BUFSIZE];
		int		bufpos;
		unsigned char	packtype;
		int		packlen;
		char		version[256];	/* self-reported version of the device (constructed from other params) */
		char		op_params[256];	/* operating parameters of the device (according to it) */
		int		manufacturer;	/* code for manufacturer specific operations */
		int		model;		/* code for specific model */

		/* general information obtained from the system */
		int		got_state;	/* flag to indicate that we've recieved the reciever states */
		unsigned char	status[3];	/* intersense status ("MS/n") */
#define MAX_UNITS 12
		int		num_units;	/* number of units that return status > 0 */
		_FastrakUnit	units[MAX_UNITS];
#define MAX_TUNITS 100				/* For polhemus usually 1, or IS-900 several more */
		int		num_tunits;	/* number of transmitter units (soniDisc beacons for Intersense trackers) */
		_FastrakTransUnit tunits[MAX_TUNITS];	/* transmitter units */

		char		*init_command;	/* string to send to tracker during initialization */
		int		stream_mode;	/* whether stream mode should be used to get data */
		int		hemisphere;	/* code of which hemisphere to use */

		char		firmware_version[16];	/* version of system software */
		char		system_id[33];	/* system id code (eg. "IS900") */
		int		reported_format;/* data will be sent in ASCII or Binary (as reported by system) */
		int		reported_units;	/* location data will be sent in Inches or Centemeters (as reported by system) */
		int		reported_mode;	/* data will be streamed or polled (as reported by system) */
		int		update_rate;	/* Hz of reported "Update rate per station" [IS900 pg 118] */
		char		genlock_id;	/* state of Genlock operation ('G', 'X' or ' ')

		/* other possibilities: */
		/*	- genlock synch: 3 integers (?) */
		/*	- ultrasonic timeout: 1 integer */
		/*	- ultrasonic sensitivity: 1 integer */
		/*	- fixed PSE: ... */

		/* information about the current values */

		int		timer;		/* incoming timer info  TODO: or is it a checksum?*/
		int		button_change;	/* boolean indicator if button values have changed*/
		int		valuator_change;/* boolean indicator of change in valuator values */
		int		receiver_change;/* boolean indicator of change in receiver values */

		/* information about how to filter the data for use with the VR library (FreeVR or CAVE) */
		float		scale_trans;	/* multiplier to scale from the Fastrak units (inches) */

#if defined(FREEVR)
		int		num_buttons;	/* the number of buttons in the configuration */
		vr2switch	**button_inputs;/* An array of pointers to the input structures */
		int		*button_map_unit;/* An array of unit numbers from which the data should come */
		int		*button_map_button;/* An array of numbers indicating the particular button */
		int		*button_values;	/* An array containing the current value of each (used for debouncing) */

		int		num_valuators;	/* the number of valuators in the configuration */
		vrValuator	**valuator_inputs;/* An array of pointers to the input structures */
		int		*valuator_map_unit;/* An array of unit numbers from which the data should come */
		int		*valuator_map_valuator;/* An array of numbers indicating the particular valuator */
		float		*valuator_sign;/* An array of numbers indicating the particular valuator */
		int		*valuator_values;	/* An array containing the current value of each (used for debouncing) */

		int		num_6sensors;	/* the number of 6-dof input sensors in config */
		vr6sensor	**sensor6_inputs;/* An array of pointers to the input structures */
		int		*sensor6_map;	/* An array of unit numbers from which the data should come */
#endif

	} _FastrakPrivateInfo;



	/******************************************************/
	/*** General NON public Fastrak interface routines ***/
	/******************************************************/

/******************************************************/
/* Since this routine is generally called before we know what type of */
/*   device we're connected to, we'll just do a generic intialization.*/
/*   (At one point I considered having the initialization be affected */
/*   by the device type, but I think that's backwards  -- although,   */
/*   now that I think about it further, it is possible to get the type*/
/*   from the configuration information.  I guess we'll leave it this */
/*   way for now.                                                     */
static void _FastrakInitializeStruct(_FastrakPrivateInfo *aux)
{
	int	count;

	aux->got_state = 0;

	aux->version[0] = '\0';
	aux->op_params[0] = '\0';
	aux->manufacturer = 0;

	aux->buf[0] = '\0';
#ifndef COMM_DEBUG
	memset(aux->buf, '\0', BUFSIZE);
#else /* debugging stuff */
	memset(aux->buf, '*', BUFSIZE);
	aux->buf[BUFSIZE-1] = '\0';
#endif
	aux->bufpos = 0;

	aux->timer = 0;
	aux->button_change = 0;
	aux->valuator_change = 0;
	aux->receiver_change = 0;

	/* TODO: set default stream_mode & filter mode here */
	aux->hemisphere = FASTRAK_SET_HEMISPHERE_LOWER;/* use lower hemisphere by default */
	aux->scale_trans = 1.0/12.0;		/* convert from inches to feet by default */

	/* initialize individual units */
	for (count = 0; count < MAX_UNITS; count++) {
		aux->units[count].tracking_status = '-';
	}

	aux->num_units = 0;
	aux->num_tunits = 0;

	/* everything else is zero'd by default */

}


/******************************************************/
static void _FastrakPrintStruct(FILE *file, _FastrakPrivateInfo *aux, vrPrintStyle style)
{
	int	count;

	vrFprintf(file, "Fastrak protocol device internal structure (0x%p):\n", aux);
	vrFprintf(file, "\r\tversion -- '%s'\n", aux->version);
	vrFprintf(file, "\r\toperating parameters -- '%s'\n", aux->op_params);
	vrFprintf(file, "\r\tfd = %d\n\tport = '%s'\n\tbaud = %d (%d)\n\topen = %d\n",
		aux->fd,
		aux->port,
		aux->baud_int, aux->baud_enum,
		aux->open);
	vrFprintf(file, "\r\tfirmware -- '%s'\n", aux->firmware_version);
	vrFprintf(file, "\r\tsystem id -- '%s'\n", aux->system_id);
	vrFprintf(file, "\r\tmanufacturer -- '%d'\n", aux->manufacturer);
	vrFprintf(file, "\r\treport format -- '%s' (%d)\n",
		(aux->reported_format == 0 ? "ASCII" : "Binary"),
		aux->reported_format);
	vrFprintf(file, "\r\treport units -- '%s' (%d)\n",
		(aux->reported_units == 0 ? "Inches" : "Centemeters"),
		aux->reported_units);
	vrFprintf(file, "\r\treport mode -- '%s' (%d)\n",
		(aux->reported_mode == 0 ? "Polled" : "Continuous"),
		aux->reported_mode);
	vrFprintf(file, "\r\tupdate rate -- %dHz\n", aux->update_rate);

	vrFprintf(file, "\r\tgot receiver states -- %d\n", aux->got_state);
	vrFprintf(file, "\r\tactive units -- %d\n", aux->num_units);
	vrFprintf(file, "\r\tactive transmitter units -- %d\n", aux->num_tunits);


#ifdef FREEVR /* { */
	vrFprintf(file, "\r\t%d buttons:\n", aux->num_buttons);
	vrFprintf(file, "\r\tbutton_values = 0x%p, button_map_unit = 0x%p, button_map_button = 0x%p, button_inputs = 0x%p\n",
		aux->button_values,
		aux->button_map_unit,
		aux->button_map_button,
		aux->button_inputs);
	for (count = 0; count < aux->num_buttons; count++) {
		vrFprintf(file, "\r\t\tbutton %d: value = %d, map_unit = %d, map_button = %d, input = 0x%p (type = %d)\n",
			count,
			aux->button_values[count],
			aux->button_map_unit[count],
			aux->button_map_button[count],
			aux->button_inputs[count],
			aux->button_inputs[count]->input_type);
	}

	vrFprintf(file, "\r\t%d valuators:\n", aux->num_valuators);
	vrFprintf(file, "\r\tvaluator_values = 0x%p, valuator_map_unit = 0x%p, valuator_map_valuator = 0x%p, valuator_inputs = 0x%p\n",
		aux->valuator_values,
		aux->valuator_map_unit,
		aux->valuator_map_valuator,
		aux->valuator_inputs);
	for (count = 0; count < aux->num_valuators; count++) {
		vrFprintf(file, "\r\t\tvaluator %d: value = %d, map_unit = %d, map_valuator = %d, input = 0x%p (type = %d)\n",
			count,
			aux->valuator_values[count],
			aux->valuator_map_unit[count],
			aux->valuator_map_valuator[count],
			aux->valuator_inputs[count],
			aux->valuator_inputs[count]->input_type);
	}

	vrFprintf(file, "\r\t%d 6-sensors:\n", aux->num_6sensors);
	vrFprintf(file, "\r\tsensor6_values = 0x%p, sensor6_map = 0x%p, sensor6_inputs = 0x%p\n",
#if 1 /* TODO: change this to a "1" -- not sure why it's 0 (copied from vr_input.asc_fob.c) */
		aux->sensor6_inputs,
#else
		0x0,
#endif
		aux->sensor6_map,
		aux->sensor6_inputs);
	for (count = 0; count < aux->num_6sensors; count++) {
		vrFprintf(file, "\r\t\t6-sensor %d: value = [%.2f %.2f %.2f  %.2f %.2f %.2f], map_unit = %d, input = 0x%p (type = %d)\n",
			count,
			aux->sensor6_inputs[count]->position->v[VR_X],
			aux->sensor6_inputs[count]->position->v[VR_Y],
			aux->sensor6_inputs[count]->position->v[VR_Z],
			aux->sensor6_inputs[count]->position->v[VR_AZIM+3],
			aux->sensor6_inputs[count]->position->v[VR_ELEV+3],
			aux->sensor6_inputs[count]->position->v[VR_ROLL+3],
			aux->sensor6_map[count],
			aux->sensor6_inputs[count],
			aux->sensor6_inputs[count]->input_type);
	}

#  if 0 /* TODO: not sure if we want/need this */
	vrFprintf(file, "\r\tcontrol inputs:\n");
	for (count = 0; count < MAX_CONTROLS; count++)
		vrFprintf(file, "\r\t\tcontrol_inputs[%d] = 0x%p\n", count, aux->control_inputs[count]);
#  endif
#endif /* } FREEVR */

	vrFprintf(file, "\r\tscale_trans = %f\n", aux->scale_trans);
	vrFprintf(file, "\r\themisphere = %d\n", aux->hemisphere);
	/* TODO: print some info about other current values */

	/* print some info about each receiver unit */
	vrFprintf(file, "\r\t%d position receivers:\n", aux->num_units);
	for (count = 0; count < aux->num_units; count++) {
		vrFprintf(file, "\t\treceiver unit %d: state = %d, toreport (%d bytes) = '%s'\n", count+1, aux->units[count].state, aux->units[count].toreport_len, aux->units[count].toreport);
	}

	/* print some info about each transmitter unit */
	vrFprintf(file, "\r\t%d position transmitters:\n", aux->num_tunits);
	for (count = 0; count < aux->num_tunits; count++) {
		vrFprintf(file, "\t\ttransmitter unit %d: pse_id = %d, loc = %7.4f,%7.4f,%7.4f, norm = %.2f,%.2f,%.2f, hw_id = %d\n", \
			count+1,
			aux->tunits[count].pse_id,
			aux->tunits[count].position[VR_X],
			aux->tunits[count].position[VR_Y],
			aux->tunits[count].position[VR_Z],
			aux->tunits[count].position[VR_X+3],
			aux->tunits[count].position[VR_Y+3],
			aux->tunits[count].position[VR_Z+3],
			aux->tunits[count].hardware_id);
	}
}


/**************************************************************************/
static void _FastrakPrintHelp(FILE *file, _FastrakPrivateInfo *aux)
{
	vrFprintf(file, BOLD_TEXT "Sorry, Fastrak - print_help control not yet implemented.\n" NORM_TEXT);
}


/******************************************************/
static char _FastrakSendCommand(_FastrakPrivateInfo *aux, _FastrakCommand command)
{
	_FastrakCommandInfo	*command_info;
	char			return_code = '\0';

	/* if command doesn't exist, then return immediately */
	if (command >= FASTRAK_LASTCOMMAND) {
		vrPrintf(RED_TEXT "No command (%d), nothing sent to Fastrak protocol device\n" NORM_TEXT, (int)command);
		return (return_code);
	}

	command_info = &(_FastrakMsgList[command]);

#ifdef COMM_DEBUG
	vrPrintf(
#else /* ) */
	vrDbgPrintfN(FASTRK_DBGLVL + RARE_DBGLVL,
#endif
		"FASTRAK: Sending command '%s'\n", command_info->name);

	vrSerialWrite(aux->fd, command_info->msg, command_info->len);
#if 1
	vrSerialDrain(aux->fd);
fsync(aux->fd);
#endif
	vrSleep(command_info->us_delay);

	/* TODO: should we check for error status after every command? */
	/*   Well, we certainly shouldn't do it after a stream command. */
	/*   Probably best to require specific code to do error checking. */

#if 0
vrPrintf("sendcommand complete: command %d ('%c') -- \"%s\"\n", (int) command, command_info->msg[0], command_info->msg);
#endif
	return (return_code);
}


/******************************************************/
static char _FastrakSendCommandAddr(_FastrakPrivateInfo *aux, _FastrakCommand command, int addr)
{
static	char			newMsg[32];
	_FastrakCommandInfo	*command_info;
	char			return_code = '\0';

	/* if command doesn't exist, then return immediately */
	if (command >= FASTRAK_LASTCOMMAND) {
		vrPrintf(RED_TEXT "No command (%d), nothing sent to Fastrak protocol device\n" NORM_TEXT, (int)command);
		return (return_code);
	}

	command_info = &(_FastrakMsgList[command]);

	if (addr < 1 || addr > 12) {
		vrPrintf(RED_TEXT "Address out of range for command '%s' (%d), nothing sent to Fastrak protocol device\n" NORM_TEXT, command_info->msg, (int)command);
		return (return_code);
	}

#ifdef COMM_DEBUG
	vrPrintf(
#else /* ) */
	vrDbgPrintfN(FASTRK_DBGLVL + RARE_DBGLVL,
#endif
		"FASTRAK: Sending command '%s' to unit #%d\n", command_info->name, addr);

	memcpy(newMsg, command_info->msg, command_info->len);

	/* place the address into the command string */
	newMsg[1] = STATION_NUM2CHAR((unsigned char)addr);

	/* put a terminating character at the end of the string */
	/* NOTE: this is only necessary for making printouts nice for debuging */
	newMsg[command_info->len] = '\0';

	vrSerialWrite(aux->fd, newMsg, command_info->len);
#if 1
	vrSerialDrain(aux->fd);
fsync(aux->fd);
#endif
	vrSleep(command_info->us_delay);

	/* TODO: should we check for error status after every command? */
	/*   Well, we certainly shouldn't do it after a stream command. */
	/*   Probably best to require specific code to do error checking. */

#if 0
vrPrintf("sendcommandaddr complete: command %d to addr %d ('%c','%c') -- \"%s\"\n", (int) command, addr, newMsg[0], newMsg[1], newMsg);
#endif
	return (return_code);
}


/******************************************************/
static char _FastrakSendCommandString(_FastrakPrivateInfo *aux, char *command)
{
	char			return_code = '\0';

	vrDbgPrintfN(
#ifdef COMM_DEBUG
		1,
#else
		FASTRK_DBGLVL + RARE_DBGLVL,
#endif
		"FASTRAK: Sending command '%s'\n", command);

	vrSerialWrite(aux->fd, command, strlen(command));
#if 1
	vrSerialDrain(aux->fd);
fsync(aux->fd);
#endif

	/* TODO: should we check for error status after every command? */
	/*   Well, we certainly shouldn't do it after a stream command. */
	/*   Probably best to require specific code to do error checking. */

	return (return_code);
}


/***********************************************************************/
/* NOTE: num_chars is only used for ASCII decoding */
float _FastrakReadFloat(unsigned char *buf, int num_chars, int binary)
{
static	char	local[10] = "\0\0\0\0\0\0\0\0\0\0";
	float	return_value;

	if (binary) {
		/* TODO: not implemented yet */
		printf("Warning, binary parsing not implemented yet\n");
		return 0.0;
	} else {
		if (num_chars > 10) {
			vrPrintf("Warning, too many characters to _FastrakReadFloat: %d\n", num_chars);
			return 0.0;
		}

		memcpy(local, buf, num_chars);
		return_value = atof(local);
	}

#if 0
	vrPrintf("_FastrakReadFloat: returning atof(%s)\n", local);
#endif
	return return_value;
}


/***********************************************************************/
/* NOTE: num_chars is only used for ASCII decoding */
int _FastrakReadIntN(unsigned char *buf, int num_chars, int binary)
{
static	char	local[10] = "\0\0\0\0\0\0\0\0\0\0";

	if (binary) {
		/* TODO: not implemented yet */
		printf("Warning, binary parsing not implemented yet\n");
		return 0;
	} else {
		if (num_chars > 10) {
			vrPrintf("Warning, too many characters to _FastrakReadIntN: %d\n", num_chars);
			return 0;
		}

		memcpy(local, buf, num_chars);
		local[num_chars] = '\0';
#if 0
	vrPrintf("_FastrakReadIntN: returning vrAtoI(%s)\n", local);
#endif
		return vrAtoI(local);
	}
}


/******************************************************/
/* _FastrakDecodeRecord() ...                         */
/*                                                    */
/* Overview of records (based on IS-900 manual):                 */
/*    (2nd byte always represents station number, and            */
/*       when reporting a system-wide value is '1').             */
/*   .	0.  -- incoming sensor data [pg. 109]                    */
/*   .	21S -- System Status and Version info [pg. 110]          */
/*   .	2.O -- Output List Record [pg. 111]                      */
/*   .	2.l -- Station State Record (on/off) [pg. 111]           */
/*	2.A -- Alignment Reference Record (NIY) [pg. 112]        */
/*	2.G -- Boresight Reference Angles Record (NIY) [pg. 112] */
/*	2.H -- Hemisphere Record (NIY) [pg. 113]                 */
/*	2.N -- Tip Offset Record (NIY) [pg. 113]                 */
/*	2.V -- Location Operational Envelope (NIY) [pg. 114]     */
/*      2 E -- An error message (NOTE: 2nd value is a space)     */
/*   .	31S -- Manufacturer System Status Record [pg. 114]       */
/*   .	3.s -- Manufacturer Station Record [pg. 115]             */
/*	3.p -- Prediction Interval Record (NIY) [pg. 116]        */
/*	3.Q -- Sensitivity Level Reocrd (NIY) [pg. 116]          */
/*	31G -- Genlock Synchronization Record (NIY) [pg. 116]    */
/*	31U -- Ultrasonic Timeout Record (NIY) [pg. 117]         */
/*	31g -- Ultrasonic Sensitivity Record (NIY) [pg. 117]     */
/*	31F -- Fixed PSE Record (NIY) [pg. 117]                  */
/*   .	31P -- Tracking Status Record [pg. 118]                  */
/*      51T -- System Test Data record (NIY)                     */
/*                                                               */
/* Here are some records that might come from a Patriot device:  */
/*   [NOTE: I've mostly limited these to ones that are possible  */
/*   given the commands that have been implemented.]             */
/*    	0.u -- Units that are enabled (only up to 2 on a Patriot)*/
/*      0.v -- Version information (00v can be 1 of 2 things)    */
/*      0.O -- Output List Record (like Fastrak 2.O)             */
/*      0.F -- Output Format (ASCII or Binary)                   */
/*      0.U -- Output Units (Inches or Centimeters)              */
/*****************************************************************/
static int _FastrakDecodeRecord(_FastrakPrivateInfo *aux)
{
	int		station_num;		/* number of the station of incoming report */
	int		shift_bytes;		/* number of bytes to shift after the decoding */

/* NOTE: the buffer is stored in aux and the data in the buffer   */
/*   is shifted after decoded, so the next decode cycle can work  */
/*   from what is now in at the beginning of the buffer.  This    */
/*   function returns a zero when there isn't a complete record   */
/*   to parse, and one when something was parsed, and there *may* */
/*   be more to parse.                                            */

	/* if there aren't enough bytes for proper decoding, then return, to come back later */
	/* NOTE: some command-responses use the second byte to determine the type of information */
	/*   that is being returned, and some use the third.  For now we will require that there */
	/*   be at least three bytes in the buffer before we do any parsing, but if there are    */
	/*   commands that will only return two bytes and no more, then there is a small chance  */
	/*   that this could cause a problem.  But we'll go with it for now.                     */
	if (aux->bufpos < 3) {	/* for robustness testing, change to "< 1" vs. "< 3" */
#ifdef COMM_DEBUG
static	int	warning_count = 0;
		vrPrintf("Skipping decoding of buffer with %d bytes\n", aux->bufpos);
		warning_count++;
		if (warning_count > 5)
			abort();
#endif
		return 0;
	}

#if defined(COMM_DEBUG) || 0
vrPrintf("Decoding buffer with %d bytes: '%s'\n", aux->bufpos, aux->buf);
#endif

	switch(aux->buf[0]) {

	/***************/
	/* Data Record */
	case '0': {
		int		count;
		int		bufpos;		/* report position */
		int		report_len;	/* length of data report */
		_FastrakUnit	in_unit;	/* place into which to decode incoming data */
		_FastrakUnit	*unit;		/* pointer to actual unit data memory */

		station_num = STATION_CHAR2NUM(aux->buf[1]);
		if (((station_num < 1) || (station_num > 12)) && !(aux->model == MODEL_PATRIOT))  {
			vrPrintf(RED_TEXT "Fastrak: Warning: bad station number: %d\n" NORM_TEXT, station_num);
			vrPrintf("Data (bufpos = %d): '%s'\n", aux->bufpos, aux->buf);

			/* mark the next record as bad, and therefore skip to newline */
			aux->buf[0] = '_';
			return 0;
		}

		/*******************************************************************/
		/* NOTE: the Polhemus Patriot also uses the "0" report for items   */
		/*   other than the data record.  So need to check for those here. */
		if (aux->model == MODEL_PATRIOT) {
#if defined(COMM_DEBUG) || 0
			vrPrintf(RED_TEXT "First checking whether '%c' is a valid patriot response.\n" NORM_TEXT, aux->buf[2]);
#endif
			switch (aux->buf[2]) {

			/* A space indicates we've got a normal data record, so fall through */
			case ' ':
				break;

			/* reporting the Format: ASCII = 0, Binary = 1 (acutally, if it's in */
			/*   Binary mode, we get Binary feedback -- which makes it difficult */
			/*   to easily parse).                                               */
			case 'F':
vrPrintf("Patriot: got an 'F' command\n");
				/* TODO: implement this */
				shift_bytes = 8;
				break;

			/* reporting the Units: Inches = 0, Centimeters = 1 */
			case 'U':
vrPrintf("Patriot: got a 'U' command\n");
				/* TODO: implement this */
				shift_bytes = 8;
				break;

			/* reporting the version information */
			/* WARNING: the return protocol for the Patriot had redundant codes   */
			/*   that report different information depending on the command given */
			/*   requesting the report.                                           */
			case 'v':
				/* TODO: implement this */
				vrPrintf(RED_TEXT "Skipping Patriot version information '%s' (skipping 128 bytes)\n" NORM_TEXT, &(aux->buf[7]));
				shift_bytes = 128;
				break;

			case 'u': {
				int		count;
				int		max_units;	/* the "l*" command only returns 32 units */
				_FastrakUnit	*unit;

				station_num = STATION_CHAR2NUM(aux->buf[1]);
				unit = &(aux->units[station_num-1]);

#ifdef COMM_DEBUG
				vrPrintf("Got a Station State Record for all stations '%s'\n", aux->buf);
#endif
				/* clamp number of units to 2 */
				max_units = MAX_UNITS;
				if (max_units > 2)
					max_units = 2;

				for (count = 0; count < max_units; count++) {
					aux->units[count].state = (int)(aux->buf[count+5] == '1');
vrPrintf("Patriot: unit %d state is %d ('%c')\n", count, aux->units[count].state, aux->buf[count+5]);
				}
				shift_bytes = 9;

				aux->got_state = 1;	/* flag that we've got the reciever states */
vrPrintf("Patriot: done checking unit states\n");

				} break;

			case 'O': {
				int		num_types = 0;
				int		pos;
				int		record_length = 4;	/* 3 bytes for the station number */
				_FastrakUnit	*unit;

				station_num = STATION_CHAR2NUM(aux->buf[1]);
				unit = &(aux->units[station_num-1]);

#if defined(COMM_DEBUG) || 1
				vrPrintf("Got Output List Record for Patriot station %d: '%s'\n", station_num, aux->buf);
#endif

				for (pos = 4; (aux->buf[pos+1] != '\r') && (pos < aux->bufpos); pos += 3) {
					int	type;

vrPrintf("Evaluating for '%c' '%c' '%c'\n", aux->buf[pos], aux->buf[pos+1], aux->buf[pos+2]);
					type = _FastrakReadIntN(&aux->buf[pos], 3, 0);

					switch (type) {
					case  0: record_length += 1;	break;	/* ASCII space character */
					case  1: record_length += 2;	break;	/* CR/LF */
					/* TODO: binary values for below -- these work for ASCII only */
					case  2: record_length += 9*3;	break;	/* x,y,z location: 3 floats */
					case  3: record_length += 15*3;	break;	/* x,y,z location: 3 high-res floats */
					case  4: record_length += 9*3;	break;	/* yaw,pitch,roll angles: 3 floats */
					case  5: record_length += 15*3;	break;	/* yaw,pitch,roll angles: 3 high-res floats */
					case  6: record_length += 87;	break;	/* Direction Cosine 3x3 matrix */
					case  7: record_length += 9*4;	break;	/* Quaternion Orientation */
					case  8: record_length += 10;	break;	/* Timestamp */
					case  9: record_length += 10;	break;	/* Framecount */
					case 10: record_length += 3;	break;	/* stylus switch status (always 0) */
					default:
						vrPrintf("Patriot: Warning, unimplemented output record type: %d for station %d\n", type, station_num);
					}
vrPrintf("Patriot output list for station %d: got type %d, record_length is now %d\n", station_num, type, record_length);
					unit->toreport_type[num_types] = type;
					num_types++;
				}
				unit->toreport_num = num_types;
				unit->toreport_len = record_length;
#if defined(COMM_DEBUG)
				vrPrintf("Patriot unit %d: report_len set to %d\n", station_num, record_length);
#endif

				memcpy(unit->toreport, &aux->buf[3], pos-3);
				unit->toreport[pos-3] = '\0';	/* terminate the string */

				pos += 3;	/* skip the space/CR/LF */
				shift_bytes = pos;
				if (aux->buf[shift_bytes-1] != '\n') {
					aux->buf[0] = '_';		/* the underscore is a signal to skip-to-NL in the next decode cycle */
					shift_bytes = 0;
				}
				} break;

			}

			if ((station_num < 1) || (station_num > 2))  {
				vrPrintf(RED_TEXT "Patriot: Warning: bad station number: %d\n" NORM_TEXT, station_num);
				vrPrintf("Data (bufpos = %d): '%s'\n", aux->bufpos, aux->buf);

				/* mark the next record as bad, and therefore skip to newline */
				aux->buf[0] = '_';
				return 0;
			}

			/* for all the non-fall-through codes, skip out of the current switch statement */
			/* Ugly, I know. */
			if (aux->buf[2] != ' ')
				break;
		}
		/* End of special Patriot section.                                 */
		/*******************************************************************/

		unit = &(aux->units[station_num-1]);	/* TODO: read data into a temporary unit location and copy at the end -- but first copy current unit values into temporary space */
		in_unit = *unit;

		report_len = in_unit.toreport_len;
#if defined(COMM_DEBUG) || 0
		vrPrintf("unit %d: report_len is %d\n", station_num, report_len);
#endif
		/* this shouldn't happen, but if report length is bigger */
		/*   than the buffer, we have some serious problems.     */
		if ((report_len > BUFSIZE || report_len < 1)) {
static	int	warning_count = 0;
			vrPrintf(RED_TEXT "Fastrak: Warning: improbable report length for unit %d: %d\n" NORM_TEXT, station_num, report_len);
			warning_count++;
			if (warning_count > 5)
				abort();

			aux->buf[0] = '_';		/* the underscore is a signal to skip-to-NL in the next decode cycle */
			return 0;
		}

		/* if there aren't enough bytes for this report, then return from the function */
		if (aux->bufpos < report_len) {
#ifdef COMM_DEBUG
			vrPrintf("Not enough data (%d) for station %d (%d bytes needed)\n", aux->bufpos, station_num, report_len);
#endif
			return 0;
		}

#ifdef COMM_DEBUG
		vrDbgPrintfN(FASTRK_DBGLVL, "Decoding Data Record (station %d): '%s' (bufpos = %d)\n", station_num, aux->buf, aux->bufpos);
#endif

		if (aux->model != MODEL_PATRIOT) {
			for (count = 0, bufpos = 3; count < in_unit.toreport_num; count++) {
				switch (in_unit.toreport_type[count]) {
#if 1
				case 0: bufpos++; break;/* just a space */
				case 1: bufpos+=2;break;/* just a CR/LF */
#else
				case 0:	break;		/* just a space */
				case 1: break;		/* just a CR/LF */
#endif
				case 2:			/* x,y,z locations in 3 floats */
					in_unit.data_pos[VR_X] = _FastrakReadFloat(&aux->buf[bufpos], 7, aux->reported_format);
					bufpos += 7;
					in_unit.data_pos[VR_Y] = _FastrakReadFloat(&aux->buf[bufpos], 7, aux->reported_format);
					bufpos += 7;
					in_unit.data_pos[VR_Z] = _FastrakReadFloat(&aux->buf[bufpos], 7, aux->reported_format);
					bufpos += 7;
					aux->receiver_change = 1;	/* TODO: ideally this would be a per-unit setting */
					break;
				case 4:			/* yaw, pitch & roll angles */
					in_unit.data_pos[VR_AZIM+3] = _FastrakReadFloat(&aux->buf[bufpos], 7, aux->reported_format);
					bufpos += 7;
					in_unit.data_pos[VR_ELEV+3] = _FastrakReadFloat(&aux->buf[bufpos], 7, aux->reported_format);
					bufpos += 7;	/* TODO: check this number */
					in_unit.data_pos[VR_ROLL+3] = _FastrakReadFloat(&aux->buf[bufpos], 7, aux->reported_format);
					bufpos += 7;	/* TODO: check this number */
					aux->receiver_change = 1;	/* TODO: ideally this would be a per-unit setting */
					break;
				case 22:		/* button information */
					/* TODO: we should be able to handle buttons for more than one wand -- ie. ideally s/b in_unit.buttons = ... 8/29/05 -- starting to address this */
					in_unit.buttons = _FastrakReadIntN(&aux->buf[bufpos], 4, aux->reported_format);
					bufpos += 5;	/* NOTE: the number is 4 ascii bytes, but for some reason it comes back with a trailing space */
					aux->button_change = 1;
					break;
				case 23:		/* joystick information */
					/* TODO: we should be able to handle joysticks for more than one wand -- ie. ideally s/b in_unit.joystick[n] = ... 8/29/05 -- starting to address this */
					in_unit.joystick[0] = _FastrakReadIntN(&aux->buf[bufpos], 4, aux->reported_format);
					bufpos += 4;
					in_unit.joystick[1] = _FastrakReadIntN(&aux->buf[bufpos], 4, aux->reported_format);
					bufpos += 4;
					aux->valuator_change = 1;
					break;
				case 40:		/* tracking quality information */
					in_unit.tracking_quality = _FastrakReadIntN(&aux->buf[bufpos], 4, aux->reported_format);
					bufpos += 4;
					break;

				case 5:			/* X-axis direction cosines */
				case 6:			/* Y-axis direction cosines */
				case 7:			/* Z-axis direction cosines */
				case 11:		/* Quaternion Orientation */
				case 18:		/* special 16 bit binary x,y,z format */
				case 19:		/* special 16 bit binary a,e,r format */
				case 20:		/* special 16 bit binary quaternion format */
				case 21:		/* time stamp data */

				case 68:		/* "auxilary input" */
				case 69:		/* "auxilary input" */
				case 70:		/* "auxilary input" */
				case 71:		/* "auxilary input" */
				default:
					vrPrintf("Warning, unimplemented data type from tracker %d -- buf = '%c'\n", in_unit.toreport_type[count], aux->buf[0]);
					break;
				}
			}
		} else {
			/* This section is for decoding of the Patriot data */
			for (count = 0, bufpos = 4; count < in_unit.toreport_num; count++) {
				switch (in_unit.toreport_type[count]) {
#if 1
				case 0: bufpos++; break;/* just a space */
				case 1: bufpos+=2;break;/* just a CR/LF */
#else
				case 0:	break;		/* just a space */
				case 1: break;		/* just a CR/LF */
#endif
				case 2:			/* x,y,z locations in 3 floats */
					in_unit.data_pos[VR_X] = _FastrakReadFloat(&aux->buf[bufpos], 9, aux->reported_format);
					bufpos += 9;
					in_unit.data_pos[VR_Y] = _FastrakReadFloat(&aux->buf[bufpos], 9, aux->reported_format);
					bufpos += 9;
					in_unit.data_pos[VR_Z] = _FastrakReadFloat(&aux->buf[bufpos], 9, aux->reported_format);
					bufpos += 9;
					aux->receiver_change = 1;	/* TODO: ideally this would be a per-unit setting */
					break;
				case 4:			/* yaw, pitch & roll angles */
					in_unit.data_pos[VR_AZIM+3] = _FastrakReadFloat(&aux->buf[bufpos], 9, aux->reported_format);
					bufpos += 9;
					in_unit.data_pos[VR_ELEV+3] = _FastrakReadFloat(&aux->buf[bufpos], 9, aux->reported_format);
					bufpos += 9;	/* TODO: check this number */
					in_unit.data_pos[VR_ROLL+3] = _FastrakReadFloat(&aux->buf[bufpos], 9, aux->reported_format);
					bufpos += 9;	/* TODO: check this number */
					aux->receiver_change = 1;	/* TODO: ideally this would be a per-unit setting */
					break;
				case 10:		/* button information */
					in_unit.buttons = _FastrakReadIntN(&aux->buf[bufpos], 3, aux->reported_format);
					bufpos += 3;	/* NOTE: the number is 4 ascii bytes, but for some reason it comes back with a trailing space */
					aux->button_change = 1;
					break;

				case 3:			/* High-res Cartesian location */
				case 5:			/* High-res Euler angles */
				case 6:			/* Direction Cosines matrix */
				case 7:			/* Quaternion Orientation */
				case 8:			/* time stamp data */
				case 9:			/* frame count data */

				default:
					vrPrintf("Warning, unimplemented data type from tracker %d -- buf = '%c'\n", in_unit.toreport_type[count], aux->buf[0]);
					break;
				}
			}
		}

		/* Make sure last character was a CR/LF -- perhaps this can be done in the general section */
		if (aux->buf[report_len-1] != '\n') {
			vrPrintf(RED_TEXT "Warning, station %d expecting a newline at byte %d, but got (0x%02x) instead: '%s'\n" NORM_TEXT, station_num, report_len, aux->buf[report_len-1], aux->buf);
			aux->buf[0] = '_';		/* the underscore is a signal to skip-to-NL in the next decode cycle */
			shift_bytes = 0;
		} else {
#if defined(COMM_DEBUG) || 0
			vrPrintf(BOLD_TEXT "Made a good decoding for station %d\n" NORM_TEXT, station_num);
#endif
			shift_bytes = report_len;

			/* now put the decoded data into the actual unit data structure */
			*unit = in_unit;
			unit->read_count++;
		}
		} break;

	/*****************/
	/* Status Record */
	case '2':

		switch (aux->buf[2]) {

		/***************************************************/
		/* System Status Record (w/ version info), pg. 110 */
		case 'S':
			/* byte 1: s/b '1'                                    */
			/* byte 3: Config Hex Char 0                          */
			/* byte 4: Config Hex Char 1                          */
			/* byte 5: Config Hex Char 2                          */
			/*	bit 0: output format (0=ASCII, 1=Binary)      */
			/*	bit 1: output units (0=inches, 1=centimeters) */
			/*	bit 3: transmit mode (0=Polled, 1=Continuous) */
			/* bytes 16-21: Firmware version ID                   */
			/* bytes 22-53: System Identification                 */

			/* verify that bufpos >= 55 and that aux->buf[1] = '1' */
			if (aux->bufpos < 55) {
				/* haven't read enough bytes yet */
				return 0;
			}

			if (aux->buf[1] != '1') {
				/* incorrect byte in the record */
				vrDbgPrintfN(AALWAYS_DBGLVL, "_FastrakDecodeRecord(): invalid record\n");
			}

			vrDbgPrintfN(FASTRK_DBGLVL, "Got Status Record: '%s'\n", aux->buf);
			aux->reported_format = aux->buf[5] & 0x01;	/* bit 0 of byte 5 */
			aux->reported_units  = (aux->buf[5] & 0x02) != 0;	/* bit 1 of byte 5 */ /* TODO: make 0 or 1 */
			aux->reported_mode   = (aux->buf[5] & 0x08) != 0;	/* bit 3 of byte 5 */ /* TODO: make 0 or 1 */
			sscanf((char *)&aux->buf[15], "%s %s", &aux->firmware_version, &aux->system_id);

			/* use the system id string to set the manufacturer */
			if (strncasecmp("is900", (char *)&aux->buf[22], 5) == 0) {
				aux->manufacturer = MANU_INTERSENSE;
				aux->model = MODEL_IS900;
			} else {
				aux->manufacturer = MANU_POLHEMUS;
				aux->model = MODEL_FASTRAK;	/* NOTE: Polhemsu Patriot (and perhaps Liberty) do not have the "S" command */
			}

			shift_bytes = 55;
			if (aux->buf[shift_bytes-1] != '\n') {
				aux->buf[0] = '_';		/* the underscore is a signal to skip-to-NL in the next decode cycle */
				shift_bytes = 0;
			}

			break;

		/*******************************/
		/* Output List Record, pg. 111 */
		case 'O': {
			int		num_types = 0;
			int		pos;
			int		record_length = 3;	/* 3 bytes for the station number */
			_FastrakUnit	*unit;

			station_num = STATION_CHAR2NUM(aux->buf[1]);
			unit = &(aux->units[station_num-1]);

#if defined(COMM_DEBUG) || 1
			vrPrintf("Got Output List Record for station %d: '%s'\n", station_num, aux->buf);
#endif

			for (pos = 3; aux->buf[pos] != '\r'; pos += 2) {
				int	type;

				type = _FastrakReadIntN(&aux->buf[pos], 2, 0);

				switch (type) {
				case  0: record_length += 1;	break;	/* ASCII space character */
				case  1: record_length += 2;	break;	/* CR/LF */
				/* TODO: binary values for below -- these work for ASCII only */
				case  2: record_length += 7*3;	break;	/* x,y,z location: 3 floats */
				case  4: record_length += 7*3;	break;	/* yaw,pitch,roll angles: 3 floats */
				case  5: record_length += 7*3;	break;	/* X-axis cosines: 3 floats */
				case  6: record_length += 7*3;	break;	/* Y-axis cosines: 3 floats */
				case  7: record_length += 7*3;	break;	/* Z-axis cosines: 3 floats */
				case 11: record_length += 7*4;	break;	/* Quaternion Orientation */
				case 16: record_length += 1;	break;	/* stylus switch status (always 0) */
				case 22: record_length += 5;	break;	/* button presses */
				case 23: record_length += 8;	break;	/* X & Y joystick values */
				case 40: record_length += 4;	break;	/* Tracking status */

				/* TODO: implement the rest of the types */
				case 18:	/* special 16 bit binary x,y,z format */
				case 19:	/* special 16 bit binary a,e,r format */
				case 20:	/* special 16 bit binary quaternion format */
				case 21:	/* time stamp data */

				case 68:	/* "auxilary input" */
				case 69:	/* "auxilary input" */
				case 70:	/* "auxilary input" */
				case 71:	/* "auxilary input" */
				default:
					vrPrintf("Fastrak: Warning, unimplemented output record type: %d for station %d\n", type, station_num);
				}
				unit->toreport_type[num_types] = type;
				num_types++;
			}
			unit->toreport_num = num_types;
			unit->toreport_len = record_length;
#if defined(COMM_DEBUG)
			vrPrintf("Fastrak unit %d: report_len set to %d\n", station_num, record_length);
#endif

			memcpy(unit->toreport, &aux->buf[3], pos-3);
			unit->toreport[pos-3] = '\0';	/* terminate the string */

			pos += 2;	/* skip the CR/LF */
			shift_bytes = pos;
			if (aux->buf[shift_bytes-1] != '\n') {
				aux->buf[0] = '_';		/* the underscore is a signal to skip-to-NL in the next decode cycle */
				shift_bytes = 0;
			}
			} break;

		/*********************************/
		/* Station State Record, pg. 111 */
		case 'l': { /* The response to an "l<n>\n" or "l*\n" command */
			int		count;
			int		max_units;	/* the "l*" command only returns 32 units */
			_FastrakUnit	*unit;

			/* NOTE: there are basically two possible response types:             */
			/*   1) the state of 4 units  (<n>234)                                */
			/*   2) the state of 32 units (123456789abcdefghijklmnopqrstuvw)      */
			/* If we get the first response, we will only deal with station <n>.  */
			/* If we get the second type of response, we will do all 32 stations. */
			station_num = STATION_CHAR2NUM(aux->buf[1]);
			unit = &(aux->units[station_num-1]);

			if (aux->buf[7] == '\r') {
#ifdef COMM_DEBUG
				vrPrintf("Got a Single Station State Record for station %d: '%s'\n", station_num, aux->buf);
#endif
				unit->state = (int)(aux->buf[3] == '1');
				shift_bytes = 9;
			} else {
#ifdef COMM_DEBUG
				vrPrintf("Got a Station State Record for all stations '%s'\n", aux->buf);
#endif
				/* clamp number of units to 32 */
				max_units = MAX_UNITS;
				if (max_units > 32)
					max_units = 32;

				for (count = 0; count < max_units; count++) {
					aux->units[count].state = (int)(aux->buf[count+3] == '1');
				}
				shift_bytes = 37;
			}

			aux->got_state = 1;	/* flag that we've got the reciever states */

			} break;

		/*****************************/
		/* Unimplemented 2.x Records */
		case 'A':
		case 'G':
		case 'H':
		case 'N':
		case 'V':
			/* skip buffer to next CR,LF */
			aux->buf[0] = '_';		/* the underscore is a signal to skip-to-NL in the next decode cycle */
			shift_bytes = 0;

			vrPrintf("Got Unimplemented Status Record of type '%c': '%s'\n", aux->buf[2], aux->buf);
			break;

		/************************/
		/* Unknown Record Types */
		default:
			aux->buf[0] = '_';		/* the underscore is a signal to skip-to-NL in the next decode cycle */
			shift_bytes = 0;
			vrPrintf("Got Unknown Status Record of type '%c': '%s'\n", aux->buf[2], aux->buf);
			break;
		}
		break;

	/****************************/
	/* InterSense Status Record */
	case '3':

		switch (aux->buf[2]) {

		/**********************************************/
		/* Manufacturer System Status Record, pg. 114 */
		/*   [NOTE: for now we assume "Manufacturer is Intersense.] */
		case 'S': /* The response to an "MS\n" command */
			/* TODO: verify that aux->buf[1] = '1' */

			vrPrintf("Got InterSense Status Record: '%s'\n", aux->buf);

			/* TODO: something interesting with this data -- see pg 73 of IS900_SIM_VE_Manual_v1.pdf */
			aux->buf[0] = '_'; shift_bytes = 0;	/* Temporary, until parsing code added */
			break;

		/****************************************/
		/* Manufacturer Station Record, pg. 115 */
		case 's': /* The response to an "Ms<n>\n" command */
			station_num = STATION_CHAR2NUM(aux->buf[1]);

			vrPrintf("Got InterSense Station %d Record: '%s'\n", station_num, aux->buf);

			/* TODO: something interesting with this data -- see pp 73-74 of IS900_SIM_VE_Manual_v1.pdf */
			aux->buf[0] = '_'; shift_bytes = 0;	/* Temporary, until parsing code added */
			break;

		/***********************************/
		/* Tracking Status Record, pg. 118 */
		case 'P': { /* The response to an "MP\n" command */
			int		count;
			int		max_units;
			_FastrakUnit	*unit;

			/* verify that aux->buf[1] = '1' */
			if (aux->buf[1] != '1') {
				vrPrintf(RED_TEXT "_FastrakDecodeRecord(): Warning 2nd byte of '31P' response is not '1' -- skipping record\n");
				aux->buf[0] = '_'; shift_bytes = 0;
				return 0;
			}

			if (aux->bufpos < 48) {
				/* haven't read enough bytes yet */
				return 0;
			}

#if defined(COMM_DEBUG) || 0
			vrPrintf("Parsing InterSense Tracking Status Record: '%s'\n", aux->buf);
#endif

			/* Parse the tracking status data -- see pg 76,79 of IS900_SIM_VE_Manual_v1.pdf */

			/* clamp max_units to 12 as per the max reported by the MP command */
			max_units = MAX_UNITS;
			if (max_units > 12)
				max_units = 12;

			for (count = 0; count < MAX_UNITS; count++) {
				unit = &(aux->units[count]);

				unit->tracking_status = aux->buf[count*3 + 3];

				/* tracking_meas & tracking_reject are single digit hex values */
				unit->tracking_meas   = (int)((aux->buf[count*3 + 4]) - '0');
				if (unit->tracking_meas > 9)
					unit->tracking_meas -= ('a' - '0' - 10);
				unit->tracking_reject = (int)((aux->buf[count*3 + 5]) - '0');
				if (unit->tracking_reject > 9)
					unit->tracking_reject -= ('a' - '0' - 10);
			}

			aux->update_rate = _FastrakReadIntN(&aux->buf[39], 4, 0);
			aux->genlock_id = aux->buf[45];

			shift_bytes = 48;

			} break;

		/***********************************/
		/* PSE data, pg. ? */
		case 'F': { /* The response to an "MCF\n" command */
			int			count;
			_FastrakTransUnit	new_tunit;

			/* verify that aux->buf[1] = '1' */
			if (aux->buf[1] != '1') {
				vrPrintf(RED_TEXT "_FastrakDecodeRecord(): Warning 2nd byte of '31F' response is not '1' -- skipping record\n");
				aux->buf[0] = '_'; shift_bytes = 0;
				return 0;
			}

			if (aux->bufpos < 70) {
				/* haven't read enough bytes yet */
				return 0;
			}

			new_tunit.assigned = 0;
			new_tunit.pse_id = _FastrakReadIntN(&aux->buf[3], 7, 0);
			new_tunit.position[VR_X] = _FastrakReadFloat(&aux->buf[10], 10, 0);
			new_tunit.position[VR_Y] = _FastrakReadFloat(&aux->buf[20], 10, 0);
			new_tunit.position[VR_Z] = _FastrakReadFloat(&aux->buf[30], 10, 0);
			new_tunit.position[VR_X+3] = _FastrakReadFloat(&aux->buf[40], 7, 0);
			new_tunit.position[VR_Y+3] = _FastrakReadFloat(&aux->buf[47], 7, 0);
			new_tunit.position[VR_Z+3] = _FastrakReadFloat(&aux->buf[54], 7, 0);
			new_tunit.hardware_id = _FastrakReadIntN(&aux->buf[61], 7, 0);

			/* Now search to see if this hardware_id is already in the list */
			for (count = 0; count < aux->num_tunits; count++) {
				if ((aux->tunits[count].assigned == 1) && (aux->tunits[count].hardware_id == new_tunit.hardware_id)) {
					new_tunit.assigned = 1;
					aux->tunits[count] = new_tunit;
				}
			}
			/* If not, add it to the list */
			if (!new_tunit.assigned) {
				new_tunit.assigned = 1;
				aux->tunits[aux->num_tunits] = new_tunit;
				aux->num_tunits++;
				vrDbgPrintfN(FASTRK_DBGLVL, "Fasktrak: new hardware id %d\n", new_tunit.hardware_id);
			} else {
				vrDbgPrintfN(FASTRK_DBGLVL, "Fasktrak: existing hardware id %d\n", new_tunit.hardware_id);
			}

			shift_bytes = 70;

			} break;

		/*************************/
		/* Unimplemented Records */
		case 'p':
		case 'Q':
		case 'G':
		case 'U':
		case 'g':
			/* skip buffer to next CR,LF */
			aux->buf[0] = '_';		/* the underscore is a signal to skip-to-NL in the next decode cycle */
			shift_bytes = 0;
			break;

		/************************/
		/* Unknown Record Types */
		default:
			/* skip buffer to next CR,LF */
			aux->buf[0] = '_';		/* the underscore is a signal to skip-to-NL in the next decode cycle */
			shift_bytes = 0;
			break;
		}
		break;

	/*******************/
	/* Whitespace type */
	case ' ':
	case '\n':
		/* TODO: skip entire whitespace string */
		shift_bytes = 1;
#if 0	/* setting this to '1' should cause a skip to the new newline */
		if (aux->buf[shift_bytes-1] != '\n') {
			aux->buf[0] = '_';		/* the underscore is a signal to skip-to-NL in the next decode cycle */
			shift_bytes = 0;
		}
#endif
		break;

	/****************************/
	/* Other Misc possibilities */
	case 'P':
		/* most probable input is the Patriot startup string */
		if (strncmp((char *)(&(aux->buf[0])), "Patriot Ready!", 14) == 0) {
			vrPrintf("Got the Patriot startup message - '%s'\n", aux->buf[0]);
			shift_bytes = 16;
		} else {
			/* TODO probably print a warning about an unknown message */
		} break;

	/***********************************/
	/* Skip to newline (NL, aka CR/LF) */
	case '_': {
		/* NOTE: this is not part of the official protocol, but rather is used by this  */
		/*   driver to indicate a past decoding failure, and therefore the need to skip */
		/*   over all data until (and including) the next newline.                      */
		int	find_nl;	/* counter to loop through data looking for newline */

		for (find_nl = 0; find_nl < aux->bufpos && aux->buf[find_nl] != '\n'; find_nl++);

		if (aux->buf[find_nl] == '\n') {
			shift_bytes = find_nl+1;
#if defined(COMM_DEBUG) || 1
			vrPrintf("skipping to newline (%d bytes)\n", shift_bytes);
#endif
		} else {
			/* need to wait until more data is read and there is a newline */
			shift_bytes = 0;
#if defined(COMM_DEBUG) || 1
			vrPrintf("need more data to skip to newline\n");
#endif
			return 0;	/* return 0 causes driver to read more data before doing more decoding (otherwise we're stuck in an infinite loop) */
		}
		} break;

	/***********************/
	/* Unknown Record type */
	default:
		aux->buf[0] = '_';		/* the underscore is a signal to skip-to-NL in the next decode cycle */
		shift_bytes = 0;
		break;
	}

	/* Now do the buffer shifting */
	if (shift_bytes > 0) {
		/* shift the buffer past parsed data */
#ifdef COMM_DEBUG
		printf("shifting %d bytes\n", shift_bytes);
#endif
		memmove(aux->buf, &aux->buf[shift_bytes], (aux->bufpos-shift_bytes));	/* Major bug removed from this line 9/7/2005 */
		//memmove(aux->buf, &aux->buf[shift_bytes], shift_bytes);	/* for robustness testing, use this line for Major bug */
		aux->bufpos -= shift_bytes;
#ifndef COMM_DEBUG
		aux->buf[aux->bufpos] = '\0';
#else
		aux->buf[aux->bufpos] = '%';	/* put a percent-sign in the buffer to help debug */
		vrDbgPrintfN(FASTRK_DBGLVL, "_FastrakDecodeRecord(): decoded & shifted %d bytes, now have %d bytes in buffer: '%s'\n", shift_bytes, aux->bufpos, aux->buf);
#endif
	}

	/* if the buffer has been exhausted, then nothing more to parse */
	if (aux->bufpos == 0)
		return 0;

	/* otherwise report that there may be more to parse */
	return 1;
}


/******************************************************/
static char _FastrakReadInput(_FastrakPrivateInfo *aux)
{
	int	bytes_read;
	char	return_code = '\0';

	/* Read as much data as is available, and decode whatever is then in the buffer */
	bytes_read = vrSerialRead(aux->fd, ((char *)aux->buf + aux->bufpos), sizeof(aux->buf) - aux->bufpos - 1);
	if (bytes_read == 0) {
		/* there wasn't any input ready */
		return 0;
	}
	if (bytes_read < 0) {
		vrPrintf(RED_TEXT "readinput: error -- %d reported from vrSerialRead()\n" NORM_TEXT, bytes_read);
	} else {
		aux->bufpos += bytes_read;
#ifdef COMM_DEBUG
		vrPrintf("readinput: just read %d bytes\n", bytes_read);
		aux->buf[aux->bufpos+1] = '$';	/* one space beyond the end of the existing data -- for looking at the input buffer */
#else
		aux->buf[aux->bufpos] = '\0';
#endif
	}

	/* decode data in buffer until no complete records are available */
	//while (printf("hey\n"),_FastrakDecodeRecord(aux)) {
	while (_FastrakDecodeRecord(aux)) {
	}

	return (return_code);
}


/******************************************************/
/* _FastrakInitializeDevice() is called in the OPEN phase  */
/*   of input interface -- after the types of inputs have  */
/*   been determined (during the CREATE phase).            */
static int _FastrakInitializeDevice(_FastrakPrivateInfo *aux, char *name)
{
	int		bytes_read;
	int		unit_count;

	if (aux == NULL) {
		vrErrPrintf("_FastrakInitializeDevice(): "
			RED_TEXT "Warning, no auxiliary data for Fastrak.\n" NORM_TEXT);
		return -1;
	}

	/******************************************************************************/
	/* First, verify that (some of) the underlying data structures are consistent */
	if (strcmp(_FastrakMsgList[FASTRAK_LASTCOMMAND].name, "FASTRAK_LASTCOMMAND") != 0) {
		/* NOTE: this generally happens when a new command is added to     */
		/*   the _FastrakMsgList[] array, but the corresponding enumerated */
		/*   value is not added to the _FastrakCommand type definition.    */
		vrErrPrintf(RED_TEXT "The Fastrak device source code is not consistent -- '_FastrakMsgList[]'\n" NORM_TEXT);
		exit(1);
	}

	aux->model = MODEL_FASTRAK;	/* default setting */

	/****************************************/
	/* Make initial contact with the Device */

	/* Go into "polled mode" for communicating with the device */
	vrPrintf("Fastrak Device '%s': Going into 'polled mode'\n", name);
	_FastrakSendCommand(aux, FASTRAK_POLLMODE);

	/* go into ASCII mode */
	_FastrakSendCommand(aux, FASTRAK_ASCIIMODE);

	/* go into Inches mode */
	_FastrakSendCommand(aux, FASTRAK_SET_INCHES);

	/* get system information */
	_FastrakSendCommand(aux, FASTRAK_SYSSTAT);	/* see IS900 pg 110 to decode */

	/* send any initialization commands from the FreeVR configuration */
	if (aux->init_command != NULL) {
		_FastrakSendCommandString(aux, aux->init_command);
	}

	/****************************************************************/
	/* request data from device necessary for proper initialization */
	_FastrakSendCommand(aux, FASTRAK_GET_ALLRECEIVERS);
	vrSleep(40000);				/* give time for the responses */

	/**********************************************************/
	/* parse all the incoming responses to the above requests */

	/* TODO: Since we know the "get recievers" command was the last sent, */
	/*   we continue to decode until that response has been parsed. */

	while (!aux->got_state) {
		while (_FastrakDecodeRecord(aux)) {}
		_FastrakReadInput(aux);
	}	/* get all the incoming data */

	/* now do manufacturer specific calls -- NOTE: must wait until after FASTRAK_SYSSTAT command is processed */
	if (aux->manufacturer == MANU_INTERSENSE) {
		vrPrintf("Fastrak Device '%s': " RED_TEXT "Sending Intersense IS900 specific commands\n" NORM_TEXT, name);
		_FastrakSendCommand(aux, IS900_GET_TUNITS);
		_FastrakSendCommand(aux, IS900_GET_TRACKING_STATUS);
		vrSleep(40000);				/* give extra time for IS-900 responses */

		_FastrakDecodeRecord(aux);

		if (aux->num_tunits < 0) {
			vrErrPrintf(RED_TEXT "_FastrakInitializeDevice(): Warning: no transmitter data obtained for Intersense device '%s'.\n" NORM_TEXT, name);
		}
	}

	/****************************************/
	/* Set the parameters for all the units */

	/* count the number of active units */
	aux->num_units = 0;
	for (unit_count = 0; unit_count < MAX_UNITS; unit_count++) {
		if (aux->units[unit_count].state) {
			aux->num_units++;
		}
	}

	for (unit_count = 0; unit_count < MAX_UNITS; unit_count++) {
		/* TODO: perhaps only send to units known to exist -- though shouldn't hurt to send to all */
		if (aux->units[unit_count].has_button)
			_FastrakSendCommandAddr(aux, FASTRAK_SET_OUTPUT_POSITION_ANGLES_BJ, unit_count+1);
		else	_FastrakSendCommandAddr(aux, FASTRAK_SET_OUTPUT_POSITION_ANGLES, unit_count+1);
	}

	/* make sure to request enough information from the system to properly decode input */
	/* NOTE: since we're setting the params, we could just know this in advance, but */
	/*   it's cleaner to cover even the times when we just make use of the existing  */
	/*    parameters.                                                                */
	for (unit_count = 0; unit_count < MAX_UNITS; unit_count++) {
		if (aux->units[unit_count].state) {
			_FastrakSendCommandAddr(aux, FASTRAK_GET_STATION_PARAMS, unit_count+1);
#if defined(COMM_DEBUG)
			vrPrintf("Sent request for station parameters for station %d\n", unit_count+1);
#endif
		}
	}

#if 0 /* not sure whether I want to implement this */
	/******************************************************************/
	/* Reget the current status of the system to verify it is running */
		/* TODO: this */
#endif

	/****************************************/
	/* Create the version and param strings */
		/* TODO: this */


	/*****************************/
	/* Set the system parameters */
		/* TODO: set for binary mode for example */



	/***************************************/
	/* Go into stream mode -- if specified */
	if (aux->stream_mode) {
		/* Set the report rate */
		/* TODO: is there a report-rate command? */

		/* Go into stream mode */
		/* TODO: this */
	}


	/************************/
	/* read the first input data */

	/* send request for data when in polled mode */
	if (!aux->stream_mode) {
		_FastrakSendCommand(aux, FASTRAK_POLL);
#define STATUS_POLL
#ifdef STATUS_POLL
		if (aux->manufacturer == MANU_INTERSENSE) {
			_FastrakSendCommand(aux, IS900_GET_TRACKING_STATUS);
		}
#endif
	}
	_FastrakReadInput(aux);

	return 0;
}


/******************************************************/
/* _PatriotInitializeDevice() is called in the OPEN phase  */
/*   of input interface -- after the types of inputs have  */
/*   been determined (during the CREATE phase).            */
static int _PatriotInitializeDevice(_FastrakPrivateInfo *aux, char *name)
{
	int		bytes_read;
	int		unit_count;

	if (aux == NULL) {
		vrErrPrintf("_PatriotInitializeDevice(): "
			RED_TEXT "Warning, no auxiliary data for Patriot.\n" NORM_TEXT);
		return -1;
	}

	/******************************************************************************/
	/* First, verify that (some of) the underlying data structures are consistent */
	if (strcmp(_FastrakMsgList[FASTRAK_LASTCOMMAND].name, "FASTRAK_LASTCOMMAND") != 0) {
		/* NOTE: this generally happens when a new command is added to */
		/*   the _FastrakMsgList[] array, but the corresponding enumerated */
		/*   value is not added to the _FastrakCommand type definition.    */
		vrErrPrintf(RED_TEXT "The Fastrak device source code is not consistent -- '_FastrakMsgList[]'\n" NORM_TEXT);
		return(-1);
	}

	aux->model = MODEL_PATRIOT;	/* default setting */

	/****************************************/
	/* Make initial contact with the Device */

	/* Go into "polled mode" for communicating with the device */
	vrPrintf("Fastrak Device '%s': Going into 'polled mode'\n", name);
	_FastrakSendCommand(aux, PATRIOT_POLLMODE);

	/* go into ASCII mode */
	_FastrakSendCommand(aux, PATRIOT_ASCIIMODE);

	/* go into Inches mode */
	_FastrakSendCommand(aux, PATRIOT_SET_INCHES);

	/* get system information */
	_FastrakSendCommand(aux, PATRIOT_SYSSTAT);	/* see IS900 pg 110 to decode */

	/* send any initialization commands from the FreeVR configuration */
	if (aux->init_command != NULL) {
		_FastrakSendCommandString(aux, aux->init_command);
	}

	/****************************************************************/
	/* request data from device necessary for proper initialization */
	_FastrakSendCommand(aux, PATRIOT_GET_ALLRECEIVERS);
	vrSleep(40000);				/* give time for the responses */

	/**********************************************************/
	/* parse all the incoming responses to the above requests */

	/* TODO: Since we know the "get recievers" command was the last sent, */
	/*   we continue to decode until that response has been parsed. */

vrPrintf("YO\n");
	while (!aux->got_state) {
		while (_FastrakDecodeRecord(aux)) {}
		_FastrakReadInput(aux);
	}	/* get all the incoming data */

	/****************************************/
	/* Set the parameters for all the units */

vrPrintf("HEY\n");
	/* count the number of active units */
	aux->num_units = 0;
	for (unit_count = 0; unit_count < MAX_UNITS; unit_count++) {
		if (aux->units[unit_count].state) {
			aux->num_units++;
		}
#if defined(COMM_DEBUG)
		vrPrintf("Patriot unit %d state is %d, %d units so far.\n", unit_count+1, aux->units[unit_count].state, aux->num_units);
#endif
	}

	for (unit_count = 0; unit_count < 2; unit_count++) {
		/* TODO: perhaps only send to units known to exist -- though shouldn't hurt to send to all */


		/* set type of data to report */
		if (aux->units[unit_count].has_button)
			_FastrakSendCommandAddr(aux, PATRIOT_SET_OUTPUT_POSITION_ANGLES_B, unit_count+1);
		else	_FastrakSendCommandAddr(aux, PATRIOT_SET_OUTPUT_POSITION_ANGLES, unit_count+1);

		/* set hemisphere to use */
		_FastrakSendCommandAddr(aux, aux->hemisphere, unit_count+1);

	}

	/* make sure to request enough information from the system to properly decode input */
	/* NOTE: since we're setting the params, we could just know this in advance, but */
	/*   it's cleaner to cover even the times when we just make use of the existing  */
	/*    parameters.                                                                */
	for (unit_count = 0; unit_count < MAX_UNITS; unit_count++) {
		if (aux->units[unit_count].state) {
			_FastrakSendCommandAddr(aux, FASTRAK_GET_STATION_PARAMS, unit_count+1);
#if defined(COMM_DEBUG)
			vrPrintf("Sent request for station parameters for station %d\n", unit_count+1);
#endif
		}
	}

	/****************************************/
	/* Create the version and param strings */
		/* TODO: this */


	/*****************************/
	/* Set the system parameters */
		/* TODO: set for binary mode for example */



	/***************************************/
	/* Go into stream mode -- if specified */
	if (aux->stream_mode) {
		/* Set the report rate */
		/* TODO: is there a report-rate command? */

		/* Go into stream mode */
		/* TODO: this */
	}


	/************************/
	/* read the first input data */

	/* send request for data when in polled mode */
	if (!aux->stream_mode) {
		_FastrakSendCommand(aux, FASTRAK_POLL);
	}
	_FastrakReadInput(aux);

	return 0;
}


/******************************************************/
static int _FastrakResetDevice(_FastrakPrivateInfo *aux)
{
	if (aux == NULL)
		return -1;

	/* TODO: figure out the proper reset command sequence */

}


/******************************************************/
static int _FastrakCloseDevice(_FastrakPrivateInfo *aux)
{
	if (aux == NULL)
		return -1;

	/* TODO: figure out the proper pre-close command sequence */

	/* close the serial port */
	vrSerialClose(aux->fd);
}



/******************************************************/
/* NOTE: these names are probably specific to the IS-900, and */
/*   so we need to check the name of the device before using  */
/*   particular names. */
static unsigned int _FastrakButtonValue(char *buttonname)
{
	switch (buttonname[0]) {
	case 'y':	/* yellow button */
	case 'Y':
	case '1':
		return Fastrak_BUTTONINDEX_x1x;
	case 'r':	/* red button */
	case 'R':
	case '2':
		return Fastrak_BUTTONINDEX_x2x;
	case 'g':	/* green button */
	case 'G':
	case '3':
		return Fastrak_BUTTONINDEX_x3x;
	case 'b':	/* blue button */
	case 'B':
	case '4':
		return Fastrak_BUTTONINDEX_x4x;
	case 'j':	/* joystik button */
	case 'J':
	case '5':
		return Fastrak_BUTTONINDEX_x5x;
	case 't':	/* trigger button */
	case 'T':
	case '6':
		return Fastrak_BUTTONINDEX_x6x;
	case 'u':	/* a pseudo "point-straight-up" button */
	case 'U':
	case '7':
		return Fastrak_BUTTONINDEX_x7x;
	}

	return -1;
}


/******************************************************/
/* Translate a string name of a valuator (the "instance" config) into a numeric value */
static unsigned int _FastrakValuatorValue(char *valuatorname)
{
	/* skip an opening sign character */
	if (valuatorname[0] == '-' || valuatorname[0] == '+')
		valuatorname++;

	switch (valuatorname[0]) {
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		return vrAtoI(valuatorname);
		break;
	/* Joystick, X and Y */
	case 'x':
	case 'X':
		return Fastrak_VALUATORINDEX_X;
	case 'y':
	case 'Y':
		return Fastrak_VALUATORINDEX_Y;
	default:
		return -1;
	}

	return -1;
}


	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/


#if defined(FREEVR) /* { */
/*****************************************************************/
/*** Functions for FreeVR access of Fastrak devices for input ***/
/*****************************************************************/


	/*************************************/
	/***  FreeVR NON public routines   ***/
	/*************************************/


/*********************************************************/
static void _FastrakParseArgs(_FastrakPrivateInfo *aux, char *args)
{
static	char	*hemi_choices[] = { "fore", "aft", "right", "left", "lower", "upper" };
static	int	hemi_values[] = {
			FASTRAK_SET_HEMISPHERE_FORE,
			FASTRAK_SET_HEMISPHERE_AFT,
			FASTRAK_SET_HEMISPHERE_RIGHT,
			FASTRAK_SET_HEMISPHERE_LEFT,
			FASTRAK_SET_HEMISPHERE_LOWER,
			FASTRAK_SET_HEMISPHERE_UPPER  };
	int	null_value = -1;
	float	scale_value = -1.0;

	/**************************************/
	/** Argument format: "port" "=" file **/
	/**************************************/
	vrArgParseString(args, "port", &(aux->port));

	/****************************************/
	/** Argument format: "baud" "=" number **/
	/****************************************/
	if (vrArgParseInteger(args, "baud", &(aux->baud_int))) {
		aux->baud_enum = vrSerialBaudIntToEnum(aux->baud_int);
	}

	/*****************************************************/
	/** Argument format: "command" "=" <command string> **/
	/*****************************************************/
	vrArgParseString(args, "command", &(aux->init_command));
#if 0
	vrPrintf(RED_TEXT "Fastrak: init_command is '%s'\n", aux->init_command);
#endif
	aux->init_command = vrProcessString(aux->init_command);
#if 0
	vrPrintf(RED_TEXT "Fastrak: init_command is now: '%s'\n", aux->init_command);
#endif

	/****************************************************************************************/
	/** Format: "hemisphere" "=" { "fore" | "aft" | "right" | "left" | "lower" | "upper" } **/
	/****************************************************************************************/
	vrArgParseChoiceInteger(args, "hemisphere", (int *)&(aux->hemisphere), hemi_choices, hemi_values);

	/**********************************************/
	/** Argument format: "transScale" "=" number **/
	/**********************************************/
	if (vrArgParseFloat(args, "transscale", &scale_value)) {
		aux->scale_trans = scale_value;
	}

	/*****************************************/
	/** Argument format: "scale" "=" number **/ /* (redundant with above) */
	/*****************************************/
	if (vrArgParseFloat(args, "scale", &scale_value)) {
		aux->scale_trans = scale_value;
	}


	/** TODO: other arguments to parse go here **/
}


/************************************************************/
static void _FastrakGetData(vrInputDevice *devinfo)
{
	_FastrakPrivateInfo	*aux = (_FastrakPrivateInfo *)devinfo->aux_data;
	int			count;
	float			scale_trans = aux->scale_trans;
	int			num_inputs;
	int			buttons[3];
static	int			buttons_last = 0;
	float			values[6];

	/* send request for data when in polled mode */
	if (!aux->stream_mode) {
		_FastrakSendCommand(aux, FASTRAK_POLL);
#ifdef STATUS_POLL
		if (aux->manufacturer == MANU_INTERSENSE) {
			_FastrakSendCommand(aux, IS900_GET_TRACKING_STATUS);
		}
#endif
	}
	num_inputs = _FastrakReadInput(aux);

	/*************/
	/** buttons **/
	if (aux->button_change) {
		for (count = 0; count < aux->num_buttons; count++) {
			vr2switch	*current_2switch;
			_FastrakUnit	*current_ftunit;
			int		mapped_6sensor;
			int		new_value;

			current_2switch = aux->button_inputs[count];
			current_ftunit = &(aux->units[aux->button_map_unit[count]]);
			mapped_6sensor = aux->button_map_unit[count];

			if (current_ftunit == NULL) {
				vrErrPrintf("_FastrakGetData: Attempt to read button from a non existant FastrakUnit (%d)\n", aux->button_map_unit[count]);
			} else {
				unsigned short	button_value = 1 << aux->button_map_button[count];

				new_value = ((current_ftunit->buttons & button_value) != 0);
				vrDbgPrintfN(FASTRK_DBGLVL, "Hey, current_ftunit->buttons = %02x (masked: %02x), button_map_button = %d, value = %02x, new_value = %d, old = %d\n",
					current_ftunit->buttons,
					(current_ftunit->buttons & button_value),
					aux->button_map_button[count],
					button_value,
					new_value,
					aux->button_values[count]);

				/* compare with value on file, and if changed then assign the new value to the input */
				if (new_value != aux->button_values[count]) {
					aux->button_values[count] = new_value;
					if (current_2switch != NULL) {
						switch (current_2switch->input_type) {
						case VRINPUT_BINARY:
							vrAssign2switchValue(current_2switch, new_value);
							break;
						case VRINPUT_CONTROL:
							vrCallbackInvokeDynamic(((vrControl *)(current_2switch))->callback, 1, new_value);
							break;
						default:
							vrErrPrintf(RED_TEXT "_FobGetData: Unable to handle button inputs that aren't Binary or Control inputs\n" NORM_TEXT);
							break;
						}
					}
				}
			}
		}

		aux->button_change = 0;
	}

	/***************/
	/** valuators **/
	if (aux->valuator_change) {
		for (count = 0; count < aux->num_valuators; count++) {
			vrValuator	*current_valuator;
			_FastrakUnit	*current_ftunit;
			int		mapped_6sensor;
			int		new_int_value;
			float		new_value;

			current_valuator = aux->valuator_inputs[count];
			current_ftunit = &(aux->units[aux->valuator_map_unit[count]]);
			mapped_6sensor = aux->valuator_map_unit[count];

			if (current_ftunit == NULL) {
				vrErrPrintf("_FastrakGetData(): Attempt to read valuator from a non existant FastrakUnit (%d)\n", aux->valuator_map_unit[count]);
			} else {
				new_int_value = current_ftunit->joystick[aux->valuator_map_valuator[count]];
				if (new_int_value != aux->valuator_values[count]) {
					/* we have a new value */
					aux->valuator_values[count] = new_int_value;

					new_value = ((float)(new_int_value - 127) * aux->valuator_sign[count]) / 127.0;
					if (new_value > 1.0)
						new_value = 1.0;
					if (new_value < -1.0)
						new_value = -1.0;

					vrDbgPrintfN(FASTRK_DBGLVL, "_FastrakGetData(): current_ftunit->j[x,y] = %d,%d, valuator_map:unit = %d, valuator = %d, sign = %.1f, values (old) = %d, new_value = %.3f\n",
						current_ftunit->joystick[0], current_ftunit->joystick[1],
						aux->valuator_map_unit[count],
						aux->valuator_map_valuator[count],
						aux->valuator_sign[count],
						aux->valuator_values[count],
						new_value);

					switch (aux->valuator_inputs[count]->input_type) {
					case VRINPUT_VALUATOR:
						vrAssignValuatorValue(current_valuator, new_value);
						break;
					case VRINPUT_CONTROL:
						vrCallbackInvokeDynamic(((vrControl *)(current_valuator))->callback, 1, &new_value);
						break;
					default:
						vrErrPrintf(RED_TEXT "_FastrakGetData: Unable to handle valuator inputs that aren't Floating or Control inputs\n" NORM_TEXT);
						break;
					}
				}
			}
		}

		aux->valuator_change = 0;
	}

	/***************/
	/** 6-sensors **/
	for (count = 0; count < aux->num_6sensors; count++) {
		vr6sensor	*current_6sensor;
		vrEuler		new_value;
		int		mapped_6sensor;
		vrMatrix	tmpmat;

		current_6sensor = aux->sensor6_inputs[count];
		mapped_6sensor = aux->sensor6_map[count];
		new_value.t[VR_X] = aux->units[mapped_6sensor].data_pos[VR_X] * scale_trans;
		new_value.t[VR_Y] = aux->units[mapped_6sensor].data_pos[VR_Y] * scale_trans;
		new_value.t[VR_Z] = aux->units[mapped_6sensor].data_pos[VR_Z] * scale_trans;
		new_value.r[VR_AZIM] = aux->units[mapped_6sensor].data_pos[VR_AZIM+3];
		new_value.r[VR_ELEV] = aux->units[mapped_6sensor].data_pos[VR_ELEV+3];
		new_value.r[VR_ROLL] = aux->units[mapped_6sensor].data_pos[VR_ROLL+3];

		vrAssign6sensorValue(current_6sensor, vrMatrixSetEulerAzimaxis(&tmpmat, &new_value, VR_Z), 0 /*, TODO: timestamp */);
		aux->receiver_change = 0;
	}
}


	/****************************************************************/
	/*    Function(s) for parsing Fastrak "input" declarations.    */
	/*                                                              */
	/*  These _DEVICE<type>Input() functions are called during the  */
	/*  CREATE phase of the input interface.                        */

/**************************************************************************/
static vrInputMatch _FastrakButtonInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
static	char		*whitespace = " \t\r\b\n";
	_FastrakPrivateInfo	*aux = (_FastrakPrivateInfo *)devinfo->aux_data;
	char		*button_name_instance;
	int		button_num;
	int		unit_number;
	int		button_choice;

	vrDbgPrintfN(FASTRK_DBGLVL, "_FastrakButtonInput(): instance = '%s'\n" NORM_TEXT, dti->instance);

	button_num = aux->num_buttons;
	aux->num_buttons++;
	unit_number = vrAtoI(dti->instance);
	button_name_instance = strchr(dti->instance, ',') + 1;			/* jump past comma*/

	if (button_name_instance != NULL) {
		button_name_instance += strspn(button_name_instance, whitespace);/* skip white */
		button_choice = _FastrakButtonValue(button_name_instance);
	} else {
		button_choice = -1;
	}

	/*************************************************/
	/* verify that the requested 'instance' is valid */
	if (button_choice == -1)
		vrErrPrintf("_FastrakButtonInput: Warning, button['%s'] did not match any known button\n", dti->instance);
	else if (unit_number < 1 || unit_number > MAX_UNITS)
		vrErrPrintf("_FastrakButtonInput: Warning, sensor number must be between %d and %d\n", 1, MAX_UNITS); /* TODO: determine better (more specific) last argument */
	else {
		aux->units[unit_number-1].has_button = 1;
		if (aux->button_inputs[button_num] != NULL)
			vrErrPrintf(RED_TEXT "_FastrakButtonInput: Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	}

	vrDbgPrintfN(INPUT_DBGLVL, "_FastrakButtonInput: assigned button #%d -- type %d of sensor %d\n",
		button_num, button_choice, unit_number);

	/***********************/
	/* now setup the input */
	aux->button_inputs[button_num] = (vr2switch *)input;
	aux->button_map_unit[button_num] = unit_number - 1;	/* change to 1-based unit code */
	aux->button_map_button[button_num] = button_choice;
	aux->button_values[button_num] = 0;	/* TODO: should vrAssign2switch() be used here?  probably */

	vrDbgPrintfN(INPUT_DBGLVL, "_FastrakButtonInput(): assigned button event of value 0x%02x to input pointer = 0x%p)\n",
		button_num, aux->button_inputs[button_num]);

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}


/**************************************************************************/
static vrInputMatch _FastrakValuatorInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
static	char		*whitespace = " \t\r\b\n";
        _FastrakPrivateInfo	*aux = (_FastrakPrivateInfo *)devinfo->aux_data;
	char			*valuator_name_instance;
	int			valuator_num;
	int			unit_number;
	int			valuator_choice;

	vrDbgPrintfN(FASTRK_DBGLVL, "_FastrakValuatorInput(): instance = '%s'\n" NORM_TEXT, dti->instance);

	valuator_num = aux->num_valuators;
	aux->num_valuators++;
	unit_number = vrAtoI(dti->instance);
	valuator_name_instance = strchr(dti->instance, ',') + 1;			/* jump past comma*/

	if (valuator_name_instance != NULL) {
		valuator_name_instance += strspn(valuator_name_instance, whitespace);/* skip white */
		valuator_choice = _FastrakValuatorValue(valuator_name_instance);
	} else {
		valuator_choice = -1;
	}

	/*************************************************/
	/* verify that the requested 'instance' is valid */
	if (valuator_choice == -1)
		vrErrPrintf("_FastrakButtonInput: Warning, button['%s'] did not match any known button\n", dti->instance);
	else if (unit_number < 1 || unit_number > MAX_UNITS)
		vrErrPrintf("_FastrakValuatorInput: Warning, sensor number must be between %d and %d\n", 1, MAX_UNITS); /* TODO: determine better (more specific) last argument */
	else {
		aux->units[unit_number-1].has_button = 1;
		if (aux->valuator_inputs[valuator_num] != NULL)
			vrErrPrintf(RED_TEXT "_FastrakValuatorInput: Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	}

	/***********************/
	/* now setup the input */
	aux->valuator_inputs[valuator_num] = (vrValuator *)input;
	aux->valuator_map_unit[valuator_num] = unit_number - 1;	/* change to 1-based unit code */
	aux->valuator_map_valuator[valuator_num] = valuator_choice;
	aux->valuator_values[valuator_num] = 0;	/* TODO: should vrAssign2switch() be used here?  probably */
	if (dti->instance[0] == '-')
		aux->valuator_sign[valuator_num] = -1.0;
	else	aux->valuator_sign[valuator_num] =  1.0;

	vrDbgPrintfN(INPUT_DBGLVL, "_FastrakValuatorInput(): assigned valuator event of value 0x%02x to input pointer = 0x%p)\n",
		valuator_num, aux->valuator_inputs[valuator_num]);

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}


/**************************************************************************/
static vrInputMatch _FastrakReceiverInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
        _FastrakPrivateInfo	*aux = (_FastrakPrivateInfo *)devinfo->aux_data;
	int			sensor_num;
	int			unit_number;

	sensor_num = aux->num_6sensors;
	aux->num_6sensors++;
#if 0 /* test to see whether this can be replaced by vrAtoI() */
	unit_number = atoi(dti->instance);
#else
	unit_number = vrAtoI(dti->instance);
#endif

	/*************************************************/
	/* verify that the requested 'instance' is valid */
	if (unit_number < 1)
		vrErrPrintf("_FastrakReceiverInput(): Warning, sensor number must be between %d and %d\n", 1, MAX_UNITS);
	else if (sensor_num > MAX_UNITS)
		vrErrPrintf("_FastrakReceiverInput(): Warning, sensor number must be between %d and %d\n", 1, MAX_UNITS);
	else if (aux->sensor6_inputs[sensor_num] != NULL) {
		printf("sensor_num is now %d\n", sensor_num);
		printf("aux->sensor6_inputs = %p\n", aux->sensor6_inputs);
		if (aux->sensor6_inputs[sensor_num] != NULL) {
			printf("hey\n");
			printf("type is '%s'\n", dti->type);
			printf("instance is '%s'\n", dti->instance);
			vrErrPrintf(RED_TEXT "_FastrakReceiverInput(): Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
				dti->type, dti->instance);
printf("old address = %p\n", aux->sensor6_inputs[sensor_num]);
		}
	}

	/***********************/
	/* now setup the input */
	aux->sensor6_inputs[sensor_num] = (vr6sensor *)input;
	aux->sensor6_map[sensor_num] = unit_number - 1; /* change from one based (unit_number) to zero based (sensor_num) */

	/* the vrAssign6sensorR2Exform() function does the parsing of the next token */
	vrAssign6sensorR2Exform(aux->sensor6_inputs[sensor_num], strchr(dti->instance, ','));

	vrDbgPrintfN(INPUT_DBGLVL, "_FastrakReceiverInput(): assigned 6sensor event of value 0x%02x to input pointer = 0x%p)\n",
		unit_number, aux->sensor6_inputs[sensor_num]);

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}


	/************************************************************/
	/***************** FreeVR Callback routines *****************/
	/************************************************************/

	/************************************************************/
	/*    Callbacks for controlling the Fastrak features.      */
	/*                                                          */

/************************************************************/
static void _FastrakSystemPauseToggleCallback(vrInputDevice *devinfo, int value)
{
	_FastrakPrivateInfo	*aux = (_FastrakPrivateInfo *)devinfo->aux_data;

        if (value == 0)
                return;

	devinfo->context->paused ^= 1;
	vrDbgPrintfN(INPUT_DBGLVL, "Fastrak Control: system_pause = %d.\n",
		devinfo->context->paused);
}

/************************************************************/
static void _FastrakPrintContextStructCallback(vrInputDevice *devinfo, int value)
{
        if (value == 0)
                return;

        vrFprintContext(stdout, devinfo->context, verbose);
}

/************************************************************/
static void _FastrakPrintConfigStructCallback(vrInputDevice *devinfo, int value)
{
        if (value == 0)
                return;

        vrFprintConfig(stdout, devinfo->context->config, verbose);
}

/************************************************************/
static void _FastrakPrintInputStructCallback(vrInputDevice *devinfo, int value)
{
        if (value == 0)
                return;

        vrFprintInput(stdout, devinfo->context->input, verbose);
}

/************************************************************/
static void _FastrakPrintStructCallback(vrInputDevice *devinfo, int value)
{
	_FastrakPrivateInfo	*aux = (_FastrakPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	_FastrakPrintStruct(stdout, aux, verbose);
}

/************************************************************/
static void _FastrakPrintHelpCallback(vrInputDevice *devinfo, int value)
{
	_FastrakPrivateInfo	*aux = (_FastrakPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	_FastrakPrintHelp(stdout, aux);
}



	/************************************************************/
	/*    Callbacks for interfacing with the Fastrak device.    */
	/*                                                          */


/************************************************************/
static void _FastrakCreateFunction(vrInputDevice *devinfo)
{
	/*** List of possible inputs ***/
static	vrInputFunction	_FastrakInputs[] = {
				{ "",		VRINPUT_2WAY, _FastrakButtonInput },
				{ "button",	VRINPUT_2WAY, _FastrakButtonInput },

				{ "",		VRINPUT_VALUATOR, _FastrakValuatorInput },
				{ "joy",	VRINPUT_VALUATOR, _FastrakValuatorInput },
				{ "joystick",	VRINPUT_VALUATOR, _FastrakValuatorInput },

				{ "",		VRINPUT_6SENSOR, _FastrakReceiverInput },
				{ "receiver",	VRINPUT_6SENSOR, _FastrakReceiverInput },

				{ NULL,		VRINPUT_UNKNOWN, NULL } };
	/*** List of control functions ***/
static	vrControlFunc	_FastrakControlList[] = {
				/* overall system controls */
				{ "system_pause_toggle", _FastrakSystemPauseToggleCallback },

				/* informational output controls */
				{ "print_context", _FastrakPrintContextStructCallback },
				{ "print_config", _FastrakPrintConfigStructCallback },
				{ "print_input", _FastrakPrintInputStructCallback },
				{ "print_struct", _FastrakPrintStructCallback },
				{ "print_help", _FastrakPrintHelpCallback },

		/** TODO: other callback control functions go here **/
				/* simulated 6-sensor selection controls */
				/* simulated 6-sensor manipulation controls */
				/* other controls */

				/* end of the list */
				{ NULL, NULL } };

	_FastrakPrivateInfo	*aux = NULL;

	/******************************************/
	/* allocate and initialize auxillary data */
	devinfo->aux_data = (void *)vrShmemAlloc0(sizeof(_FastrakPrivateInfo));
	aux = (_FastrakPrivateInfo *)devinfo->aux_data;
	_FastrakInitializeStruct(aux);

	/******************/
	/* handle options */
	aux->port = vrShmemStrDup(DEFAULT_PORT);	/* default, if no port given */
	aux->baud_int = DEFAULT_BAUD;			/* default, if no baud given */
	aux->baud_enum = vrSerialBaudIntToEnum(aux->baud_int);
	_FastrakParseArgs(aux, devinfo->args);

	/***************************************/
	/* create the inputs and self-controls */
	/* Because a Fastrak type device can have a large unknown number of inputs,  we allocate  */
	/*   memory for handling them here.  First the vrInputCreateDataContainers() function     */
	/*   is called to determine how many of each type of input needs to be created.  Normally */
	/*   vrInputCreateDataContainers() is called by vrInputCreateDataContainers() (see below),*/
	/*   but for this circumstance we split the two operations.                               */
	vrInputCountDataContainers(devinfo);

	aux->num_buttons = 0;
	aux->button_inputs = (vr2switch **)vrShmemAlloc0((devinfo->num_2ways + devinfo->num_scontrols) * sizeof(vr2switch *));
	aux->button_map_unit = (int *)vrShmemAlloc0((devinfo->num_2ways + devinfo->num_scontrols) * sizeof(int));
	aux->button_map_button = (int *)vrShmemAlloc0((devinfo->num_2ways + devinfo->num_scontrols) * sizeof(int));
	aux->button_values = (int *)vrShmemAlloc0((devinfo->num_2ways + devinfo->num_scontrols) * sizeof(int));

	aux->num_valuators = 0;
	aux->valuator_inputs = (vrValuator **)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(vrValuator *));
	aux->valuator_map_unit = (int *)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(int));
	aux->valuator_map_valuator = (int *)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(int));
	aux->valuator_sign = (float *)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(float));
	aux->valuator_values = (int *)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(int));

	aux->num_6sensors = 0;
	aux->sensor6_inputs = (vr6sensor **)vrShmemAlloc0(devinfo->num_6sensors * sizeof(vr6sensor *));
#if 0
	aux->sensor6_values = (vrEuler *)vrShmemAlloc0(devinfo->num_6sensors * sizeof(vrEuler));
#endif
	aux->sensor6_map = (int *)vrShmemAlloc0(devinfo->num_6sensors * sizeof(int));

	/* Here we return to the conventional way of creating the inputs */
	vrInputCreateDataContainers(devinfo, _FastrakInputs);
	vrInputCreateSelfControlContainers(devinfo, _FastrakInputs, _FastrakControlList);
	/* TODO: anything to do for NON self-controls?  Implement for Magellan first. */

	vrDbgPrintf(BOLD_TEXT "Done creating Fastrak inputs for '%s'\n" NORM_TEXT, devinfo->name);
	devinfo->created = 1;

	return;
}


/************************************************************/
static void _FastrakOpenFunction(vrInputDevice *devinfo)
{
	_FastrakPrivateInfo	*aux = (_FastrakPrivateInfo *)devinfo->aux_data;

	/*******************/
	/* open the device */
	aux->fd = vrSerialOpen(aux->port, aux->baud_enum);
	if (aux->fd < 0) {
		vrErrPrintf("(%s::_FastrakOpenFunction::%d) error: "
			RED_TEXT "couldn't open serial port %s for %s\n" NORM_TEXT,
			__FILE__, getpid(), aux->port, devinfo->name);
		aux->open = 0;
		sprintf(aux->version, "- unconnected Fastrak -");
	} else {
		if (strcasecmp(devinfo->type, "patriot") == 0) {
			if (_PatriotInitializeDevice(aux, devinfo->name) < 0) {
				vrErrPrintf("_PatriotOpenFunction(): "
					RED_TEXT "Warning, unable to initialize '%s' Patriot.\n" NORM_TEXT,
					devinfo->name);
			}
		} else if (_FastrakInitializeDevice(aux, devinfo->name) < 0) {
			vrErrPrintf("_FastrakOpenFunction(): "
				RED_TEXT "Warning, unable to initialize '%s' Fastrak.\n" NORM_TEXT,
				devinfo->name);
		}
		aux->open = 1;
	}
	vrDbgPrintf("Done opening Fastrak input device '%s'.\n", devinfo->name);
	devinfo->ready = 1;	/* TODO: s/b = aux->open */

	return;
}


/************************************************************/
static void _FastrakCloseFunction(vrInputDevice *devinfo)
{
	_FastrakPrivateInfo	*aux = (_FastrakPrivateInfo *)devinfo->aux_data;

	if (aux != NULL) {
		_FastrakCloseDevice(aux);
		vrSerialClose(aux->fd);
		vrShmemFree(aux);	/* aka devinfo->aux_data */
	}

	return;
}


/************************************************************/
static void _FastrakResetFunction(vrInputDevice *devinfo)
{
	_FastrakPrivateInfo	*aux = (_FastrakPrivateInfo *)devinfo->aux_data;

	if (aux != NULL) {
		_FastrakResetDevice(aux);
	}

	return;
}


/************************************************************/
static void _FastrakPollFunction(vrInputDevice *devinfo)
{
#if 1
	_FastrakPrivateInfo	*aux = (_FastrakPrivateInfo *)devinfo->aux_data;

	if (vrDbgDo(FASTRK_DBGLVL)) {
static		int	first_time = 1;
		if (first_time) {
			_FastrakPrintStruct(stdout, aux, verbose);
			first_time = 0;
		}
	}
#endif
	if (devinfo->ready) {
		_FastrakGetData(devinfo);
	} else {
		/* TODO: try to open the device again */
	}

	return;
}



	/******************************/
	/*** FreeVR public routines ***/
	/******************************/


/******************************************************/
void vrFastrakInitInfo(vrInputDevice *devinfo)
{
	devinfo->version = (char *)vrShmemStrDup("-get from Fastrak device-");
	devinfo->Create = vrCallbackCreateNamed("Fastrak:Create-Def", _FastrakCreateFunction, 1, devinfo);
	devinfo->Open = vrCallbackCreateNamed("Fastrak:Open-Def", _FastrakOpenFunction, 1, devinfo);
	devinfo->Close = vrCallbackCreateNamed("Fastrak:Close-Def", _FastrakCloseFunction, 1, devinfo);
	devinfo->Reset = vrCallbackCreateNamed("Fastrak:Reset-Def", _FastrakResetFunction, 1, devinfo);
	devinfo->PollData = vrCallbackCreateNamed("Fastrak:PollData-Def", _FastrakPollFunction, 1, devinfo);
	devinfo->PrintAux = vrCallbackCreateNamed("Fastrak:PrintAux-Def", _FastrakPrintStruct, 0);

	vrDbgPrintfN(INPUT_DBGLVL, "vrFastrakInitInfo: callbacks created.\n");
}


#endif /* } FREEVR */



	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/



#if defined(TEST_APP) /* { */

/******************************************************************/
/* Ugh, I hate globals, but I don't know a better way to get the  */
/*   aux value into the interrupt signal function exit_testapp(). */
static	int	done = 0;


/*******************************************************************/
void _RenderValuesLine(_FastrakPrivateInfo *aux)
{
static	int		last_rc = -1;		/* keep track of last read_count total, and don't print if same */
	int		rc_sum;			/* sum of all the receiver read counts */
	int		unit;
	_FastrakUnit	*unitdata;		/* pointer to the current unit */

	rc_sum = 0;
	for (unit = 0; unit < aux->num_units; unit++) {
		rc_sum += aux->units[unit].read_count;
	}
	if (rc_sum == last_rc) {
		vrPrintf(".");
		fflush(stdout);
		return;
	}
	last_rc = rc_sum;

	for (unit = 0; unit < aux->num_units; unit++) {
		unitdata = &(aux->units[unit]);

		if (unitdata->has_button) {
			vrPrintf("   Rec %d [%3d(%c%X%X)]: bj=%02x,%3d,%3d %6.1f %6.1f %6.1f %6.1f %6.1f %6.1f",
				unit+1,
				unitdata->/*tracking_quality,	 robustness test:*/ read_count,
				unitdata->tracking_status,
				unitdata->tracking_meas,
				unitdata->tracking_reject,
				unitdata->buttons,
				unitdata->joystick[0],
				unitdata->joystick[1],
				unitdata->data_pos[0],	/* X */
				unitdata->data_pos[1],	/* Y */
				unitdata->data_pos[2],	/* Z */
				unitdata->data_pos[3],	/* AZIM */
				unitdata->data_pos[4],	/* ELEV */
				unitdata->data_pos[5]);	/* ROLL */
		} else {
			vrPrintf("   Rec %d [%3d(%c%X%X)]: %6.1f %6.1f %6.1f %6.1f %6.1f %6.1f",
				unit+1,
				unitdata->/*tracking_quality,	 robustness test:*/ read_count,
				unitdata->tracking_status,
				unitdata->tracking_meas,
				unitdata->tracking_reject,
				unitdata->data_pos[0],	/* X */
				unitdata->data_pos[1],	/* Y */
				unitdata->data_pos[2],	/* Z */
				unitdata->data_pos[3],	/* AZIM */
				unitdata->data_pos[4],	/* ELEV */
				unitdata->data_pos[5]);	/* ROLL */
		}
	}
	vrPrintf("\n");
}


/*******************************************************************/
/* NOTE: this routine relies on using the ANSI terminal character */
/*   codes.  Since this is just a test application, I decided to  */
/*   forgo the requirement of the curses library, and instead     */
/*   limit the terminal on which this version would work.         */
/*   On Linux do "man console_codes" to get the sequence codes    */
/*   for other operations.                                        */
void _RenderValuesScreen(_FastrakPrivateInfo *aux)
{
	int		unit;
	_FastrakUnit	*unitdata;		/* pointer to the current unit */

	int		count;

	int		col_width = 15;		/* the number of characters per column */
	int		col1_offset = 5;	/* the offset before the first column */
	int		num_lines = 40;		/* number of lines for showing chart */
	int		num_columns = 100;

	float		max_trans = 96.0;	/* 8 feet in one direction */
	float		max_rot = 180.0;	/* 180 degrees max in each direction */
	float		max_qual = 255.0;	/* 0-255 is the tracking quality range */

	int		center = (num_lines-4) / 2.0 + 4;	/* shift for header */
	float		trans_scale = (num_lines-4) * -0.5 / max_trans;
	float		rot_scale = (num_lines-4) * -0.5 / max_rot;
	float		qual_scale = (num_lines-6) * -1.0 / max_qual;


	/* save current cursor position */
	vrPrintf("7");

	/* move to upper left corner of screen */
	vrPrintf("[0;0H");

	/* clear to end of screen */
	vrPrintf("[J");

	for (unit = 0; unit < aux->num_units; unit++) {
		unitdata = &(aux->units[unit]);

		vrPrintf("[4m");	/* Change color to green */

		/* move to top of column and print unit # (underlined) */
		vrPrintf("[2;%dHRec_%d:[%3d(%c%X%X)]", unit*col_width + col1_offset, unit+1,
			unitdata->/*tracking_quality*/read_count,
			unitdata->tracking_status,
			unitdata->tracking_meas,
			unitdata->tracking_reject);


#if 1
		/* print button value (if appropriate) */
		if (unitdata->has_button)
			vrPrintf("[3;%dH (%02x) ", unit*col_width + col1_offset, unitdata->buttons);
#endif

		/* now print info headers */
		vrPrintf("[4;%dHQXYZAER", unit*col_width + col1_offset);
		if (unitdata->has_button) {
			vrPrintf("xy");
		}

		/* print a *zero* line */
		vrPrintf("[%d;%dH-------", center, unit*col_width + col1_offset);
		if (unitdata->has_button) {
			vrPrintf("--");
		}

		vrPrintf("[24m");	/* Change color to default */

		/* print the Q value (tracker quality) */
		vrPrintf("[%d;%dHm", (int)(unitdata->tracking_meas*qual_scale*16)+num_lines-1, unit*col_width + col1_offset);
		vrPrintf("[%d;%dHr", (int)(unitdata->tracking_reject*qual_scale*16)+num_lines-1, unit*col_width + col1_offset);
		vrPrintf("[%d;%dHT", (int)(unitdata->tracking_quality*qual_scale)+num_lines-1, unit*col_width + col1_offset);

		/* print the X value */
		vrPrintf("[%d;%dHX", (int)(unitdata->data_pos[VR_X]*trans_scale)+center, unit*col_width + col1_offset+1);

		/* print the Y value */
		vrPrintf("[%d;%dHY", (int)(unitdata->data_pos[VR_Y]*trans_scale)+center, unit*col_width + col1_offset+2);

		/* print the Z value */
		vrPrintf("[%d;%dHZ", (int)(unitdata->data_pos[VR_Z]*trans_scale)+center, unit*col_width + col1_offset+3);

		/* print the A value */
		vrPrintf("[%d;%dHA", (int)(unitdata->data_pos[VR_AZIM+3]*rot_scale)+center, unit*col_width + col1_offset+4);

		/* print the E value */
		vrPrintf("[%d;%dHE", (int)(unitdata->data_pos[VR_ELEV+3]*rot_scale)+center, unit*col_width + col1_offset+5);

		/* print the R value */
		vrPrintf("[%d;%dHR", (int)(unitdata->data_pos[VR_ROLL+3]*rot_scale)+center, unit*col_width + col1_offset+6);

		/* print the joystick values */
		if (unitdata->has_button) {
			vrPrintf("[%d;%dHx", (int)((unitdata->joystick[VR_X]+1)*qual_scale)+num_lines-1, unit*col_width + col1_offset+7);
			vrPrintf("[%d;%dHy", (int)((unitdata->joystick[VR_Y]+1)*qual_scale)+num_lines-1, unit*col_width + col1_offset+8);
		}
#if 0
if (unit == 0) {
vrPrintf("[%d;%dH Tracking quality for unit 0 is %d, vert pos is %d = (%d * %f) + num_lines", 30, 50, unitdata->tracking_quality, (int)(unitdata->tracking_quality*qual_scale)+num_lines-1, unitdata->tracking_quality, qual_scale);
vrPrintf("[%d;%dH X value for unit 0 is %f, vert pos is %d = (%f * %f) + %d", 32, 50, unitdata->data_pos[VR_X], (int)(unitdata->data_pos[VR_X]*trans_scale)+center, unitdata->data_pos[VR_X], trans_scale, center);
vrPrintf("[%d;%dH X,Y joystick values for unit 0 are %d,%d, vert pos are %d,%d", 34, 50, unitdata->joystick[0], unitdata->joystick[1], (int)(unitdata->joystick[VR_X]*qual_scale)+num_lines-1, (int)(unitdata->joystick[VR_Y]*qual_scale)+num_lines);
}
#endif
	}

	if (0) {	/* choose to display all the fixed PSE's (or not) */
		for (count = 0; count < aux->num_tunits; count++) {
			vrPrintf("[%d;%dHbeacon %d: pse = %d, loc = %7.4f,%7.4f,%7.4f, norm = %.2f,%.2f,%.2f, id = %d", \
				8+count, 35,
				count+1,
				aux->tunits[count].pse_id,
				aux->tunits[count].position[VR_X],
				aux->tunits[count].position[VR_Y],
				aux->tunits[count].position[VR_Z],
				aux->tunits[count].position[VR_X+3],
				aux->tunits[count].position[VR_Y+3],
				aux->tunits[count].position[VR_Z+3],
				aux->tunits[count].hardware_id);
		}
	}

	if (0) {	/* choose to display a running graph of the tracking quality (or not) */
		static	int	ring1[300];		/* TODO: should have a double array */
		static	int	ring2[300];
		static	int	index = 0;
		int		col_offset = 35;	/* TODO: base this off number of recievers */
		int		num_columns = 240;
		int		col;
		int		col_div = 3;		/* number of data columns per text column */

		/* TODO: use a double array based on number of recievers */
		ring1[index] = aux->units[0].tracking_quality;
		ring2[index] = aux->units[1].tracking_quality;

		for (col = 0; col < num_columns; col++) {
			/* print the quality for receiver-1 */
			vrPrintf("[%d;%dH*", (int)(ring1[col]*qual_scale)+num_lines-1, (col/col_div)+col_offset);
			vrPrintf("[%d;%dHo", (int)(ring2[col]*qual_scale)+num_lines-1, (col/col_div)+col_offset);
		}

		/* redo the current value with numbers so we can see movement */
		vrPrintf("[%d;%dH1", (int)(ring1[index]*qual_scale)+num_lines-1, (index/col_div)+col_offset);
		vrPrintf("[%d;%dH2", (int)(ring2[index]*qual_scale)+num_lines-1, (index/col_div)+col_offset);


		index++;
		index %= num_columns;
	}

	/* return to the original cursor position */
	vrPrintf("8");
	fflush(stdout);
}


/*******************************************************************/
void exit_testapp()
{
	done = 1;
}


/*******************************************************************/
/* A test program to communicate with a Fastrak protocol device and print the results. */
main(int argc, char *argv[])
{
	_FastrakPrivateInfo	*aux;
	int			count;
	int			numevents;
	int			rendermode = 1;
	int			device_fastrak = 1;

	done = 0;
	signal(SIGINT, exit_testapp);


	/******************************/
	/* setup the device structure */
	aux = (_FastrakPrivateInfo *)malloc(sizeof(_FastrakPrivateInfo));
	memset(aux, 0, sizeof(_FastrakPrivateInfo));
	_FastrakInitializeStruct(aux);


	/****************************************************/
	/* adjust parameters based on environment variables */
	aux->port = getenv("FASTRAK_TTY");
	if (aux->port == NULL)
		aux->port = DEFAULT_PORT;		/* default, if no file given */
	aux->baud_int = DEFAULT_BAUD;			/* default, if no baud given */
	aux->baud_enum = vrSerialBaudIntToEnum(aux->baud_int);

	if (getenv("FT_LINERENDER") != NULL)
		rendermode = 0;

	if (getenv("FT_SCREENRENDER") != NULL)
		rendermode = 1;

	/* allow for testing of the patriot device */
	if (getenv("FT_PATRIOT") != NULL)
		device_fastrak = 0;


	/*****************************************************/
	/* adjust parameters based on command-line-arguments */
	/* TODO: parse CLAs for non-default serial port, baud, etc. */


	/**************************************************/
	/* open the serial port and initialize the device */
	aux->fd = vrSerialOpen(aux->port, aux->baud_enum);
	if (aux->fd < 0) {
		fprintf(stderr, RED_TEXT "couldn't open serial port %s\n" NORM_TEXT, aux->port);
		aux->open = 0;
		sprintf(aux->version, "- unconnected Fastrak -");
	} else {
		if (device_fastrak == 1) {
			/* TODO: avoid this hardcoding w/ environment variables and/or a CLI */
			aux->units[1].has_button = 1;

			if (_FastrakInitializeDevice(aux, "Fastrak-protocol") < 0) {
				vrErrPrintf("main: " RED_TEXT "Warning, unable to initialize Fastrak.\n" NORM_TEXT);
			}
		} else {
			/* TODO: avoid this hardcoding w/ environment variables and/or a CLI */
			aux->units[0].has_button = 1;

			if (_PatriotInitializeDevice(aux, "Patriot") < 0) {
				vrErrPrintf("main: " RED_TEXT "Warning, unable to initialize Patriot.\n" NORM_TEXT);
			}
		}
		aux->open = 1;
	}

	_FastrakPrintStruct(stdout, aux, verbose);


	/**********************/
	/* display the output */

	/* for screen-rendering, scroll everything up so it won't be erased */
	if (rendermode == 1) {
		for (count = 0; count < 60; count++)
			vrPrintf("\n");
vrPrintf("[%d;%dH Test message", 28, 55);
	}

	/* just loop and print values */
	while(aux->open && !done) {
		/* send request for data when in polled mode */
		if (!aux->stream_mode) {
			_FastrakSendCommand(aux, FASTRAK_POLL);
#ifdef STATUS_POLL
			if (aux->manufacturer == MANU_INTERSENSE) {
				_FastrakSendCommand(aux, IS900_GET_TRACKING_STATUS);
			}
#endif
		}

		_FastrakReadInput(aux);

		switch(rendermode) {
		default:
		case 0:
			_RenderValuesLine(aux);
			break;
		case 1:
			_RenderValuesScreen(aux);
			break;
		}
	}

	if (aux != NULL) {
		vrSerialClose(aux->fd);
		free(aux);			/* aka devinfo->aux_data */
	}


	/*****************/
	/* close up shop */
	_FastrakCloseDevice(aux);
	vrPrintf(BOLD_TEXT "\nFastrak device closed\n" NORM_TEXT);
}

#endif /* } TEST_APP */

